"""Classes for SDMX messages.

:class:`Message` and related classes are not defined in the SDMX
:doc:`information model <implementation>`, but in the
:ref:`SDMX-ML standard <formats>`.

pandaSDMX also uses :class:`DataMessage` to encapsulate SDMX-JSON data returned
by data sources.
"""
from typing import (
    List,
    Text,
    Union,
    )

from pandasdmx.model import (
    _AllDimensions,
    AgencyScheme,
    CategoryScheme,
    Codelist,
    ConceptScheme,
    ContentConstraint,
    DataSet,
    DataflowDefinition,
    DataStructureDefinition,
    Dimension,
    DimensionComponent,
    InternationalString,
    Item,
    ProvisionAgreement,
    )
from pandasdmx.util import BaseModel, DictLike, summarize_dictlike
from requests import Response
from warnings import warn
from .writer import write


def _summarize(obj, fields):
    """Helper method for __repr__ on Header and Message (sub)classes."""
    for name in fields:
        attr = getattr(obj, name)
        if attr is None:
            continue
        yield f'{name}: {attr!r}'


class Header(BaseModel):
    """Header of an SDMX-ML message.

    SDMX-JSON messages do not have headers.
    """
    #: (optional) Error code for the message.
    error: Text = None
    #: Identifier for the message.
    id: Text = None
    #: Date and time at which the message was generated.
    prepared: Text = None
    #: Intended recipient of the message, e.g. the user's name for an
    #: authenticated service.
    receiver: Text = None
    #: The :class:`.Agency` associated with the data :class:`~.source.Source`.
    sender: Union[Item, Text] = None

    def __repr__(self):
        """String representation."""
        lines = ['<Header>']
        lines.extend(_summarize(self, self.__fields__.keys()))
        return '\n  '.join(lines)


class Footer(BaseModel):
    """Footer of an SDMX-ML message.

    SDMX-JSON messages do not have footers.
    """
    #:
    severity: Text
    #: The body text of the Footer contains zero or more blocks of text.
    text: List[InternationalString] = []
    #:
    code: int


class Message(BaseModel):
    class Config:
        # for .response
        arbitrary_types_allowed = True

    #: :class:`Header` instance.
    header: Header = Header()
    #: (optional) :class:`Footer` instance.
    footer: Footer = None
    #: :class:`requests.Response` instance for the response to the HTTP request
    #: that returned the Message. This is not part of the SDMX standard.
    response: Response = None

    def to_pandas(self, *args, **kwargs):
        """Convert a Message instance to :mod:`pandas` object(s).

        :func:`pandasdmx.writer.write` is called and passed
        the `Message` instance  as first argument, followed  by any `args` and `kwargs`.

        .. seealso:: :meth:`write`
        """
        return write(self, *args, **kwargs)

    def write(self, *args, **kwargs):
        """Alias for `to_pandas` improving backwards compatibility.

        .. deprecated:: 1.0
            Use :meth:`to_pandas` instead.
        """
        warn('Message.write() is deprecated. Use Message.to_pandas() instead.',
            DeprecationWarning)
        return self.to_pandas(*args, **kwargs)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        """String representation."""
        lines = [
            f'<pandasdmx.{self.__class__.__name__}>',
            repr(self.header).replace('\n', '\n  '),
        ]
        lines.extend(_summarize(self, ['footer', 'response']))
        return '\n  '.join(lines)


class ErrorMessage(Message):
    pass


class StructureMessage(Message):
    #: Collection of :class:`.CategoryScheme`.
    category_scheme: DictLike[str, CategoryScheme] = DictLike()
    #: Collection of :class:`.Codelist`.
    codelist: DictLike[str, Codelist] = DictLike()
    #: Collection of :class:`.ConceptScheme`.
    concept_scheme: DictLike[str, ConceptScheme] = DictLike()
    #: Collection of :class:`.ContentConstraint`.
    constraint: DictLike[str, ContentConstraint] = DictLike()
    #: Collection of :class:`.DataflowDefinition`.
    dataflow: DictLike[str, DataflowDefinition] = DictLike()
    #: Collection of :class:`.DataStructureDefinition`.
    structure: DictLike[str, DataStructureDefinition] = DictLike()
    #: Collection of :class:`.AgencyScheme`.
    organisation_scheme: DictLike[str, AgencyScheme] = DictLike()
    #: Collection of :class:`.ProvisionAgreement`.
    provisionagreement: DictLike[str, ProvisionAgreement] = DictLike()

    def __repr__(self):
        """String representation."""
        lines = [super().__repr__()]

        # StructureMessage contents
        for attr in self.__dict__.values():
            if isinstance(attr, DictLike) and attr:
                lines.append(summarize_dictlike(attr))

        return '\n  '.join(lines)


class DataMessage(Message):
    """Data Message.

    .. note:: A DataMessage may contain zero or more :class:`.DataSet`, so
       :attr:`data` is a list. To retrieve the first (and possibly only)
       data set in the message, access the first element of the list:
       ``msg.data[0]``.
    """
    #: :class:`list` of :class:`.DataSet`.
    data: List[DataSet] = []
    #: :class:`.DataflowDefinition` that contains the data.
    dataflow: DataflowDefinition = DataflowDefinition()
    #: The "dimension at observation level".
    observation_dimension: Union[_AllDimensions, DimensionComponent,
                                 List[DimensionComponent]] = None

    # Convenience access
    @property
    def structure(self):
        """DataStructureDefinition used in the :attr:`dataflow`."""
        return self.dataflow.structure

    def __repr__(self):
        """String representation."""
        lines = [super().__repr__()]

        # DataMessage contents
        if self.data:
            lines.append('DataSet ({})'.format(len(self.data)))
        lines.extend(_summarize(self, ('dataflow', 'observation_dimension')))

        return '\n  '.join(lines)



from collections.abc import Mapping

import pandas as pd
import pytest

import pandasdmx as sdmx
from pandasdmx import message, model
from pandasdmx.model import Key

from . import MessageTest


class StructuredMessageTest(MessageTest):
    """Variant of MessageTest for structure-specific messages."""
    path = MessageTest.path / 'ECB_EXR'
    dsd = None

    # Fixtures
    @pytest.fixture(scope='class')
    def dsd(self):
        yield sdmx.read_sdmx(self.path / self.dsd_filename).structure[0]

    @pytest.fixture(scope='class')
    def msg(self, dsd):
        yield sdmx.read_sdmx(self.path / self.filename, dsd=dsd)

    # Tests for every class
    def test_msg(self, dsd):
        # The message can be parsed
        sdmx.read_sdmx(self.path / self.filename, dsd=dsd)

    def test_structured_by(self, dsd, msg):
        # The DSD was used to parse the message
        assert msg.data[0].structured_by is dsd


class TestStructSpecFlatDataSet(StructuredMessageTest):
    filename = 'ng-flat-ss.xml'
    dsd_filename = 'ng-structure-full.xml'

    def test_msg_type(self, msg):
        assert isinstance(msg, message.DataMessage)
        # assert msg.data[0].dim_at_obs == 'AllDimensions'

    def test_dataset_cls(self, msg):
        assert isinstance(msg.data[0], model.DataSet)

    def test_generic_obs(self, msg):
        data = msg.data[0]
        # empty series list
        assert len(list(data.series)) == 0
        obs_list = data.obs
        assert len(obs_list) == 12
        o0 = obs_list[0]

        # Flat data set → all six dimensions are at the observation level
        assert len(o0) == 6
        assert o0.key.FREQ == 'M'
        assert o0.key.CURRENCY == 'CHF'
        assert o0.value == '1.3413'

        # All attributes are at the observation level
        assert len(o0.attrib) == 7
        assert o0.attrib.OBS_STATUS == 'A'
        assert o0.attrib.DECIMALS == '4'

    def test_write2pandas(self, msg):
        data_series = sdmx.to_pandas(msg, attributes='')
        assert isinstance(data_series, pd.Series)


class TestStructSpecSeriesDataSet(StructuredMessageTest):
    filename = 'ng-ts-gf-ss.xml'
    dsd_filename = 'ng-structure-full.xml'

    def test_msg_type(self, dsd, msg):
        # assert msg.data[0].dim_at_obs == 'TIME_PERIOD'
        pass

    def test_dataset_cls(self, msg):
        assert isinstance(msg.data[0], model.DataSet)

    def test_obs(self, msg):
        data = msg.data[0]
        assert len(data.obs) == 12
        series_list = list(data.series)
        assert len(series_list) == 4
        s3 = series_list[3]
        assert isinstance(s3, model.SeriesKey)

        # Time series data set → five dimensions are at the SeriesKey level
        assert len(s3) == 5
        assert s3.CURRENCY == 'USD'

        # 5 of 7 attributes are at the Observation level
        assert len(s3.attrib) == 5
        assert s3.attrib.DECIMALS == '4'

        obs_list = data.series[s3]
        assert len(obs_list) == 3
        o0 = obs_list[2]

        # One remaining dimension is at the Observation Level
        assert len(o0.dimension) == 1
        assert o0.dim == Key(TIME_PERIOD='2010-08')
        assert o0.value == '1.3898'

        # Two remaining attributes are at the Observation level
        assert len(o0.attached_attribute)
        assert o0.attrib.OBS_STATUS == 'A'

    def test_pandas(self, msg):
        data = msg.data[0]

        # Expected number of observations and series
        assert len(data.obs) == 12
        assert len(data.series) == 4

        # Single series can be converted to pandas
        s3 = sdmx.to_pandas(data.series[3], attributes='')
        assert isinstance(s3, pd.Series)
        # With expected values
        assert s3[0] == 1.2894

        # Single series can be converted with attributes
        s3_attr = sdmx.to_pandas(data.series[3], attributes='osgd')

        # yields a DataFrame
        assert isinstance(s3_attr, pd.DataFrame)
        assert s3_attr.shape == (3, 8)

        assert s3_attr.iloc[0].value == 1.2894

        # Attributes of observations can be accessed
        assert s3_attr.iloc[0].OBS_STATUS == 'A'

    def test_write2pandas(self, msg):
        df = sdmx.to_pandas(msg, attributes='')
        assert isinstance(df, pd.Series)
        assert df.shape == (12,)
        # with metadata
        df = sdmx.to_pandas(msg, attributes='osgd')
        assert df.shape == (12, 8)
        assert df.iloc[1].OBS_STATUS == 'A'


class TestStructSpecSeriesDataSet2(StructuredMessageTest):
    filename = 'ng-ts-ss.xml'
    dsd_filename = 'ng-structure-full.xml'

    def test_msg_type(self, dsd, msg):
        # assert msg.data[0].dim_at_obs == 'TIME_PERIOD'
        pass

    def test_dataset_cls(self, msg):
        assert isinstance(msg.data[0], model.DataSet)

    def test_structured_obs(self, msg):
        data = msg.data[0]

        # Expected number of observations and series
        assert len(data.obs) == 12
        assert len(data.series) == 4

        # SeriesKey is accessible by index using DictLike
        s3 = list(data.series.keys())[3]
        assert isinstance(s3, model.SeriesKey)

        # SeriesKey has expected number of dimensions and values
        assert len(s3) == 5
        assert s3.CURRENCY == 'USD'

        # SeriesKey has expected number of attributes and values
        assert len(s3.attrib) == 5
        assert s3.attrib.DECIMALS == '4'

        # Series observations can be accessed
        obs_list = data.series[s3]
        assert len(obs_list) == 3
        o0 = obs_list[2]
        assert len(o0.dimension) == 1
        assert o0.dim == Key(TIME_PERIOD='2010-08')
        assert o0.value == '1.3898'
        assert o0.attrib.OBS_STATUS == 'A'

    def test_dataframe(self, msg):
        data = msg.data[0]
        s = sdmx.to_pandas(data, attributes='')
        assert isinstance(s, pd.Series)
        assert len(s) == 12


class TestStructSpecSeriesData_SiblingGroup_TS(StructuredMessageTest):
    filename = 'sg-ts-ss.xml'
    dsd_filename = 'sg-structure.xml'

    def test_groups(self, msg):
        data = msg.data[0]
        assert len(data.group) == 4
        assert len(data.series) == 4
        g2 = list(data.group.keys())[2]
        assert g2.CURRENCY == 'JPY'
        print(list(data.group.keys()), g2, g2.attrib, sep='\n')
        assert g2.attrib.TITLE == \
            'ECB reference exchange rate, Japanese yen/Euro'
        # Check group attributes of a series
        s = list(data.series)[0]
        g_attrib = s.group_attrib
        assert isinstance(g_attrib, Mapping)
        assert len(g_attrib) == 1


class TestStructSpecSeriesData_RateGroup_TS(StructuredMessageTest):
    filename = 'rg-ts-ss.xml'
    dsd_filename = 'rg-structure.xml'

    def test_groups(self, msg):
        data = msg.data[0]
        assert len(data.group) == 5
        assert len(data.series) == 4
        g2 = list(data.group.keys())[2]
        assert g2.CURRENCY == 'GBP'
        assert g2.attrib.TITLE == \
            'ECB reference exchange rate, U.K. Pound sterling /Euro'
        # Check group attributes of a series
        s = list(data.series)[0]
        g_attrib = s.group_attrib
        assert isinstance(g_attrib, Mapping)
        assert len(g_attrib) == 5



"""Tests against the actual APIs for specific data sources.

HTTP responses from the data sources are cached in tests/data/cache.
To force the data to be retrieved over the Internet, delete this directory.
"""
# TODO add a pytest argument for clearing this cache in conftest.py
import logging
import os

from pandasdmx.api import Request
from pandasdmx.exceptions import HTTPError
from pandasdmx.source import DataContentType, sources
from pandasdmx.util import Resource
import pytest
import requests_mock

from .data import BASE_PATH as TEST_DATA_PATH


log = logging.getLogger(__name__)


pytestmark = pytest.mark.skipif(
    # Default value in get() ensures that when *not* on Travis, the tests run
    condition=os.environ.get('TRAVIS_EVENT_TYPE', 'cron') != 'cron',
    # For development/debugging, uncomment the following to *always* run
    # condition=False,
    reason="Fragile source tests only run on Travis for 'cron' events.")


structure_endpoints = list(filter(lambda r: r != Resource.data, Resource))


def pytest_generate_tests(metafunc):
    """pytest hook for parametrizing tests with 'endpoint' arguments."""
    if 'endpoint' not in metafunc.fixturenames:
        return  # Don't need to parametrize this metafunc

    endpoints = []

    # SDMX-JSON sources do not support structure queries
    source = sources[metafunc.cls.source_id]
    if source.data_content_type == DataContentType.JSON:
        metafunc.parametrize('endpoint', endpoints)
        return

    # This exception is raised by api.Request._request_from_args
    # TODO parametrize force=True to query these endpoints anyway; then CI
    #      XPASS will reveal when data sources change their support for
    #      endpoints
    mark_unsupported = pytest.mark.xfail(
        strict=True,
        reason='Known non-supported endpoint.',
        raises=NotImplementedError)

    for ep in structure_endpoints:
        # Accumulate multiple marks; first takes precedence
        marks = []

        # Check if the associated source supports the endpoint
        supported = source.supports[ep]
        if not supported:
            marks.append(mark_unsupported)

        # Check if the test function's class contains an expected failure
        # for this endpoint
        exc_class = metafunc.cls.xfail.get(ep.name, None)
        if exc_class:
            # Mark the test as expected to fail
            marks.append(pytest.mark.xfail(strict=True, raises=exc_class))

            if not supported:
                log.info(f'tests for {metafunc.cls.source_id!r} mention '
                         f'unsupported endpoint {ep.name!r}')

        # Tolerate 503 errors
        if metafunc.cls.tolerate_503:
            marks.append(
                pytest.mark.xfail(
                    raises=HTTPError,
                    reason='503 Server Error: Service Unavailable')
                )

        endpoints.append(pytest.param(ep, marks=marks))

    # Run the test function once for each endpoint
    metafunc.parametrize('endpoint', endpoints)


class DataSourceTest:
    """Base class for data source tests."""
    # TODO also test data endpoints
    # TODO also test structure-specific data

    # Must be one of the IDs in sources.json
    source_id = None

    # Mapping of endpoint → Exception subclass.
    # Tests of these endpoints are expected to fail.
    xfail = {}

    # True to xfail if a 503 Error is returned
    tolerate_503 = False

    @pytest.fixture
    def req(self):
        # Use a common cache file for all agency tests
        (TEST_DATA_PATH / '.cache').mkdir(exist_ok=True)
        self._cache_path = TEST_DATA_PATH / '.cache' / self.source_id
        return Request(self.source_id, cache_name=str(self._cache_path),
                       backend='sqlite')

    @pytest.mark.remote_data
    def test_endpoints(self, req, endpoint):
        # See pytest_generate_tests() for values of 'endpoint'
        cache = self._cache_path.with_suffix('.' + endpoint)
        result = req.get(endpoint, tofile=cache)

        # For debugging
        # print(cache)
        # print(cache.read_text())
        # print(result)
        # assert False

        del result


class TestABS(DataSourceTest):
    source_id = 'ABS'


class TestECB(DataSourceTest):
    source_id = 'ECB'
    xfail = {
        # 404 Client Error: Not Found
        'provisionagreement': HTTPError,
        }


# Data for requests_mock; see TestESTAT.mock()
estat_mock = {
    ('http://ec.europa.eu/eurostat/SDMX/diss-web/rest/data/nama_10_gdp/'
     '..B1GQ+P3.'): {
        'body': TEST_DATA_PATH / 'ESTAT' / 'footer2.xml',
        'headers': {
            'Content-Type':
                'application/vnd.sdmx.genericdata+xml; version=2.1',
            },
        },
    'http://ec.europa.eu/eurostat/SDMX/diss-web/file/7JUdWyAy4fmjBSWT': {
        # This file is a trimmed version of the actual response for the above
        # query
        'body': TEST_DATA_PATH / 'ESTAT' / 'footer2.zip',
        'headers': {
            'Content-Type': 'application/octet-stream',
            },
        },
    }


class TestESTAT(DataSourceTest):
    source_id = 'ESTAT'
    xfail = {
        # 404 Client Error: Not Found
        # NOTE the ESTAT service does not give a general response that contains
        #      all datastructures; this is really more of a 501.
        'datastructure': HTTPError,
        }

    @pytest.fixture
    def mock(self):
        # Prepare the mock requests
        fixture = requests_mock.Mocker()
        for url, args in estat_mock.items():
            # str() here is for Python 3.5 compatibility
            args['body'] = open(str(args['body']), 'rb')
            fixture.get(url, **args)

        return fixture

    def test_xml_footer(self, mock):
        req = Request(self.source_id)

        with mock:
            msg = req.get(url=list(estat_mock.keys())[0],
                          get_footer_url=(1, 1))

        assert len(msg.data[0].obs) == 43

    @pytest.mark.remote_data
    def test_ss_data(self, req):
        """Test a request for structure-specific data.

        Examples from:
        https://ec.europa.eu/eurostat/web/sdmx-web-services/example-queries
        """
        df_id = 'nama_10_gdp'
        args = dict(resource_id=df_id)

        # Query for the DSD
        dsd = req.dataflow(**args).dataflow[df_id].structure

        # Even with ?references=all, ESTAT returns a short message with the
        # DSD as an external reference. Query again to get its actual contents.
        if dsd.is_external_reference:
            dsd = req.get(resource=dsd).structure[0]
            print(dsd)

        assert not dsd.is_external_reference

        # Example query, using the DSD already retrieved
        args.update(dict(
            key=dict(UNIT=['CP_MEUR'], NA_ITEM=['B1GQ'], GEO=['LU']),
            params={'startPeriod': '2012', 'endPeriod': '2015'},
            dsd=dsd,
            # commented: for debugging
            # tofile='temp.xml',
        ))
        req.data(**args)


class TestIMF(DataSourceTest):
    source_id = 'IMF'


class TestILO(DataSourceTest):
    source_id = 'ILO'

    xfail = {
        # 413 Client Error: Request Entity Too Large
        'codelist': HTTPError,
        }

    @pytest.mark.remote_data
    def test_codelist(self, req):
        req.get('codelist', 'CL_ECO',
                tofile=self._cache_path.with_suffix('.' + 'codelist-CL_ECO'))


@pytest.mark.xfail(reason='500 Server Error returned for all requests.',
                   raises=HTTPError)
class TestINEGI(DataSourceTest):
    source_id = 'INEGI'

    @pytest.mark.remote_data
    def test_endpoints(self, req, endpoint):
        # SSL certificate verification sometimes fails for this server; works
        # in Google Chrome
        req.session.verify = False

        # Otherwise identical
        super().test_endpoints(req, endpoint)


class TestINSEE(DataSourceTest):
    source_id = 'INSEE'

    tolerate_503 = True

    @pytest.mark.remote_data
    def test_endpoints(self, req, endpoint):
        # Using the default 'INSEE' agency in the URL gives a response "La
        # syntaxe de la requête est invalide."
        req.get(endpoint, provider='all',
                tofile=self._cache_path.with_suffix('.' + endpoint))

@pytest.mark.xfail(reason='Service is currently unavailable.',
                   raises=HTTPError)
class TestISTAT(DataSourceTest):
    source_id = 'ISTAT'

    @pytest.mark.remote_data
    def test_gh_75(self, req):
        """Test of https://github.com/dr-leo/pandaSDMX/pull/75."""

        df_id = '47_850'

        # Reported Dataflow query works
        df = req.dataflow(df_id).dataflow[df_id]

        # dict() key for the query
        data_key = dict(
            FREQ=['A'],
            ITTER107=['001001'],
            SETTITOLARE=['1'],
            TIPO_DATO=['AUTP'],
            TIPO_GESTIONE=['ALL'],
            TIPSERVSOC=['ALL'],
        )

        # Dimension components are in the correct order
        assert [dim.id for dim in df.structure.dimensions.components] == \
            list(data_key.keys()) + ['TIME_PERIOD']

        # Reported data query works
        req.data(df_id, key='A.001001+001002.1.AUTP.ALL.ALL')

        # Use a dict() key to force Request to make a sub-query for the DSD
        req.data(df_id, key=data_key)


class TestNB(DataSourceTest):
    source_id = 'NB'
    # This source returns a valid SDMX Error message (100 No Results Found)
    # for the 'categoryscheme' endpoint.


class TestOECD(DataSourceTest):
    source_id = 'OECD'


class TestSGR(DataSourceTest):
    source_id = 'SGR'


class TestUNESCO(DataSourceTest):
    source_id = 'UNESCO'
    xfail = {
        # Requires registration
        'categoryscheme': HTTPError,
        'codelist': HTTPError,
        'conceptscheme': HTTPError,
        'dataflow': HTTPError,
        'provisionagreement': HTTPError,

        # Because 'supports_series_keys_only' was set
        # TODO check
        # 'datastructure': NotImplementedError,
        }


class TestUNSD(DataSourceTest):
    source_id = 'UNSD'


class TestWB(DataSourceTest):
    source_id = 'WB'



from time import sleep
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

import requests
from requests.compat import urlparse

from . import Source as BaseSource


class Source(BaseSource):
    """Handle Eurostat's mechanism for large datasets.

    For some requests, ESTAT returns a DataMessage that has no content except
    for a ``<footer:Footer>`` element containing a URL where the data will be
    made available as a ZIP file.

    To configure :meth:`finish_message`, pass its `get_footer_url` argument to
    :meth:`pandasdmx.api.Request.get`.

    .. versionadded:: 0.2.1

    """
    _id = 'ESTAT'

    def modify_request_args(self, kwargs):
        super().modify_request_args(kwargs)

        kwargs.pop('get_footer_url', None)

    def finish_message(self, message, request, get_footer_url=(30, 3),
                       **kwargs):
        """Handle the initial response.

        This hook identifies the URL in the footer of the initial response,
        makes a second request (polling as indicated by *get_footer_url*), and
        returns a new DataMessage with the parsed content.

        Parameters
        ----------
        get_footer_url : (int, int)
            Tuple of the form (`seconds`, `attempts`), controlling the interval
            between attempts to retrieve the data from the URL, and the
            maximum number of attempts to make.
        """
        # Check the message footer for a text element that is a valid URL
        url = None
        for text in getattr(message.footer, 'text', []):
            if urlparse(str(text)).scheme:
                url = str(text)
                break

        if not url:
            return message

        # Unpack arguments
        wait_seconds, attempts = get_footer_url

        # Create a temporary file to store the ZIP response
        ntf = NamedTemporaryFile(prefix='pandasdmx-')
        # Make a limited number of attempts to retrieve the file
        for a in range(attempts):
            sleep(wait_seconds)
            try:
                # This line succeeds if the file exists; the ZIP response
                # is stored to ntf, and then used by the
                # handle_response() hook below
                return request.get(url=url, tofile=ntf)
            except requests.HTTPError:
                raise
        ntf.close()
        raise RuntimeError('Maximum attempts exceeded')

    def handle_response(self, response, content):
        """Handle the polled response.

        The request for the indicated ZIP file URL returns an octet-stream;
        this handler saves it, opens it, and returns the content of the single
        contained XML file.

        """
        
        if response.headers['content-type'] != 'application/octet-stream':
            return response, content

        # Read all the input, forcing it to be copied to
        # content.tee_filename
        while True:
            if len(content.read()) == 0:
                break

        # Open the zip archive
        with ZipFile(content.tee, mode='r') as zf:
            # The archive should contain only one file
            infolist = zf.infolist()
            assert len(infolist) == 1

            # Set the new content type
            response.headers['content-type'] = 'application/xml'

            # Use the unzipped archive member as the response content
            return response, zf.open(infolist[0])



from . import Source as BaseSource


class Source(BaseSource):
    _id = 'ISTAT'

    def modify_request_args(self, kwargs):
        """Supply explicit provider agency ID for ISTAT.

        As reported by @guglielmo, https://github.com/dr-leo/pandaSDMX/pull/75,
        ISTAT requires a provider agency ID for structure endpoints. Values
        'IT1' and 'all' are known to work; we use 'all' to (hopefully) be
        inclusive of 'IT1' and any others.
        """
        super().modify_request_args(kwargs)

        # NB this is an indirect test for resource_type != 'data'; because of
        #    the way the hook is called, resource_type is not available
        #    directly.
        if 'key' not in kwargs:
            kwargs.setdefault('provider', 'all')



from enum import Enum
from typing import (
    TYPE_CHECKING,
    KT,
    VT,
    Any,
    Union,
    Type,
    TypeVar,
    get_type_hints,
    no_type_check,
    )
try:
    from typing import OrderedDict
except ImportError:
    # Python < 3.7.2 compatibility; see
    # https://github.com/python/cpython/commit/68b56d0
    import collections
    from typing import _alias
    OrderedDict = _alias(collections.OrderedDict, (KT, VT))


import pydantic
from pydantic import DictError, Extra, ValidationError
from pydantic.class_validators import make_generic_validator


class Resource(str, Enum):
    """Enumeration of SDMX REST API endpoints.

    ====================== ================================================
    :class:`Enum` member   :mod:`pandasdmx.model` class
    ====================== ================================================
    ``categoryscheme``     :class:`CategoryScheme \
                                   <pandasdmx.model.CategoryScheme>`
    ``codelist``           :class:`Codelist \
                                   <pandasdmx.model.Codelist>`
    ``conceptscheme``      :class:`ConceptScheme \
                                   <pandasdmx.model.ConceptScheme>`
    ``data``               :class:`DataSet \
                                   <pandasdmx.model.DataSet>`
    ``dataflow``           :class:`DataflowDefinition \
                                   <pandasdmx.model.DataflowDefinition>`
    ``datastructure``      :class:`DataStructureDefinition \
                                   <pandasdmx.model.DataStructureDefinition>`
    ``provisionagreement`` :class:`ProvisionAgreement \
                                   <pandasdmx.model.ProvisionAgreement>`
    ====================== ================================================
    """

    # agencyscheme = 'agencyscheme'
    # attachementconstraint = 'attachementconstraint'
    # categorisation = 'categorisation'
    categoryscheme = 'categoryscheme'
    codelist = 'codelist'
    conceptscheme = 'conceptscheme'
    # contentconstraint = 'contentconstraint'
    data = 'data'
    # dataconsumerscheme = 'dataconsumerscheme'
    dataflow = 'dataflow'
    # dataproviderscheme = 'dataproviderscheme'
    datastructure = 'datastructure'
    # hierarchicalcodelist = 'hierarchicalcodelist'
    # metadata = 'metadata'
    # metadataflow = 'metadataflow'
    # metadatastructure = 'metadatastructure'
    # organisationscheme = 'organisationscheme'
    # organisationunitscheme = 'organisationunitscheme'
    # process = 'process'
    provisionagreement = 'provisionagreement'
    # reportingtaxonomy = 'reportingtaxonomy'
    # schema = 'schema'
    # structure = 'structure'
    # structureset = 'structureset'

    @classmethod
    def from_obj(cls, obj):
        """Return an enumeration value based on the class of *obj*."""
        clsname = {
            'DataStructureDefinition': 'datastructure',
            }.get(obj.__class__.__name__, obj.__class__.__name__)
        return cls[clsname.lower()]

    @classmethod
    def describe(cls):
        return '{' + ' '.join(v.name for v in cls._member_map_.values()) + '}'


if TYPE_CHECKING:
    Model = TypeVar('Model', bound='BaseModel')


class BaseModel(pydantic.BaseModel):
    """Shim for pydantic.BaseModel.

    This class changes two behaviours in pydantic. The methods are direct
    copies from pydantic's code, with marked changes.

    1. https://github.com/samuelcolvin/pydantic/issues/524

       - "Multiple RecursionErrors with self-referencing models"
       - In e.g. pandasdmx.model.Item, having both .parent and .child
         references leads to infinite recursion during validation.
       - Fix: override BaseModel.__setattr__.
       - New value 'limited' for Config.validate_assignment: no sibling
         field values are passed to Field.validate().
       - New key Config.validate_assignment_exclude: list of field names that
         are not validated per se *and* not passed to Field.validate() when
         validating a sibling field.
    2. https://github.com/samuelcolvin/pydantic/issues/521

       - "Assignment to attribute changes id() but not referenced object,"
         marked as wontfix by pydantic maintainer.
       - When cls.attr is typed as BaseModel (or a subclass), then
         a.attr is b.attr is always False, even when set to the same reference.
       - Fix: override BaseModel.validate() without copy().
    """
    class Config:
        validate_assignment = 'limited'
        validate_assignment_exclude = []

    # Workaround for https://github.com/samuelcolvin/pydantic/issues/521
    @classmethod
    def validate(cls: Type['Model'], value: Any) -> 'Model':
        if isinstance(value, dict):
            return cls(**value)
        elif isinstance(value, cls):
            return value  # ***
        elif cls.__config__.orm_mode:
            return cls.from_orm(value)
        else:
            try:
                value_as_dict = dict(value)
            except (TypeError, ValueError) as e:
                raise DictError() from e
            return cls(**value_as_dict)

    # Workaround for https://github.com/samuelcolvin/pydantic/issues/524
    @no_type_check
    def __setattr__(self, name, value):
        if (self.__config__.extra is not Extra.allow and name not in
                self.__fields__):
            raise ValueError(f'"{self.__class__.__name__}" object has no field'
                             f' "{name}"')
        elif not self.__config__.allow_mutation:
            raise TypeError(f'"{self.__class__.__name__}" is immutable and '
                            'does not support item assignment')
        elif (self.__config__.validate_assignment and name not in
              self.__config__.validate_assignment_exclude):
            if self.__config__.validate_assignment == 'limited':
                kw = {'include': {}}
            else:
                kw = {'exclude': {name}}
            known_field = self.__fields__.get(name, None)
            if known_field:
                value, error_ = known_field.validate(value, self.dict(**kw),
                                                     loc=name)
                if error_:
                    raise ValidationError([error_], type(self))
        self.__dict__[name] = value
        self.__fields_set__.add(name)


def get_class_hint(obj, attr):
    """Return the type hint for attribute *attr* on *obj*."""
    klass = get_type_hints(obj.__class__)[attr].__args__[0]
    if getattr(klass, '__origin__', None) is Union:
        klass = klass.__args__[0]
    return klass


class DictLike(OrderedDict[KT, VT]):
    """Container with features of a dict & list, plus attribute access."""
    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            if isinstance(key, int):
                return list(self.values())[key]
            elif isinstance(key, str) and key.startswith('__'):
                raise AttributeError
            else:
                raise

    def __setitem__(self, key, value):
        key = self._apply_validators('key', key)
        value = self._apply_validators('value', value)
        super().__setitem__(key, value)

    # Access items as attributes
    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError as e:
            raise AttributeError(*e.args) from None

    def validate(cls, value, field):
        if not isinstance(value, (dict, DictLike)):
            raise ValueError(value)

        result = DictLike()
        result.__fields = {'key': field.key_field, 'value': field}
        result.update(value)
        return result

    def _apply_validators(self, which, value):
        try:
            field = self.__fields[which]
        except AttributeError:
            return value
        result, error = field._apply_validators(
            value, validators=field.validators, values={}, loc=(), cls=None)
        if error:
            raise ValidationError([error], self.__class__)
        else:
            return result


def summarize_dictlike(dl, maxwidth=72):
    """Return a string summary of the DictLike contents."""
    value_cls = dl[0].__class__.__name__
    count = len(dl)
    keys = ' '.join(dl.keys())
    result = f'{value_cls} ({count}): {keys}'

    if len(result) > maxwidth:
        # Truncate the list of keys
        result = result[:maxwidth - 3] + '...'

    return result


def validate_dictlike(*fields):
    def decorator(cls):
        v = make_generic_validator(DictLike.validate)
        for field in fields:
            cls.__fields__[field].post_validators = [v]
        return cls

    return decorator



"""Test code appearing in the documentation.

The 'remote_data' mark, from the pytest plugin 'pytest-remotedata', is used to
mark tests that will access the Internet. In order to run these tests, a
command-line argument must be given:

$ py.test --remote-data [...]

"""
import numpy as np
import pandas as pd
import pytest

import pandasdmx as sdmx
from pandasdmx import Request
from pandasdmx.model import GenericDataSet
from pandasdmx.util import DictLike

from . import assert_pd_equal
from .data import specimen


@pytest.mark.remote_data
def test_doc_example():
    """Code from example.rst."""
    import pandasdmx as sdmx
    estat = sdmx.Request('ESTAT')

    metadata = estat.datastructure('DSD_une_rt_a')

    for cl in 'CL_AGE', 'CL_UNIT':
        print(sdmx.to_pandas(metadata.codelist[cl]))

    resp = estat.data(
        'une_rt_a',
        key={'GEO': 'EL+ES+IE'},
        params={'startPeriod': '2007'},
        )

    data = sdmx.to_pandas(resp) \
               .xs('Y15-74', level='AGE', drop_level=False)

    data.loc[('A', 'Y15-74', 'PC_ACT', 'T')]

    # Further checks per https://github.com/dr-leo/pandaSDMX/issues/157

    # DimensionDescriptor for the structure message
    dd1 = metadata.structure.DSD_une_rt_a.dimensions

    # DimensionDescriptor retrieved whilst validating the data message
    dd2 = resp.data[0].structured_by.dimensions

    # DimensionDescriptors have same ID, components and order
    assert dd1 == dd2

    # One SeriesKey from the data message
    sk = list(resp.data[0].series.keys())[0]

    # Key values have same order as in the DSD
    assert dd1.order_key(sk) == sk


@pytest.mark.remote_data
def test_doc_index1():
    """First code example in index.rst."""
    estat = Request('ESTAT')
    flow_response = estat.dataflow('une_rt_a')

    with pytest.raises(TypeError):
        # This presumes the DataStructureDefinition instance can conduct a
        # network request for its own content
        structure_response = flow_response.dataflow.une_rt_a.structure(
            request=True, target_only=False)

    # Same effect
    structure_response = estat.get(
        'datastructure', flow_response.dataflow.une_rt_a.structure.id)

    # Even better: Request.get(…) should examine the class and ID of the object
    # structure = estat.get(flow_response.dataflow.une_rt_a.structure)

    # Show some codelists
    s = sdmx.to_pandas(structure_response)
    expected = pd.Series({
        'AT': 'Austria',
        'BE': 'Belgium',
        'BG': 'Bulgaria',
        'CH': 'Switzerland',
        'CY': 'Cyprus',
        }, name='GEO') \
        .rename_axis('CL_GEO')

    # Codelists are converted to a DictLike
    assert isinstance(s.codelist, DictLike)

    # Same effect
    assert_pd_equal(s.codelist['CL_GEO'].sort_index().head(), expected)


@pytest.mark.remote_data
def test_doc_usage_structure():
    """Code examples in walkthrough.rst."""
    ecb = Request('ECB')

    ecb_via_proxy = Request('ECB', proxies={'http': 'http://1.2.3.4:5678'})
    assert all(getattr(ecb_via_proxy.session, k) == v for k, v in (
        ('proxies', {'http': 'http://1.2.3.4:5678'}),
        ('stream', False),
        ('timeout', 30.1),
        ))

    msg1 = ecb.categoryscheme()

    assert msg1.response.url == (
        'http://sdw-wsrest.ecb.int/service/categoryscheme/ECB/latest'
        '?references=parentsandsiblings')

    # Check specific headers
    headers = msg1.response.headers
    assert headers['Content-Type'] == ('application/vnd.sdmx.structure+xml; '
                                       'version=2.1')
    assert all(k in headers for k in ['Connection', 'Date', 'Server'])

    # Removed: in pandaSDMX 0.x this was a convenience method that (for this
    # structure message) returned two DataStructureDefinitions. Contra the
    # spec, that assumes:
    # - There is 1 Categorization using the CategoryScheme; there could be
    #   many.
    # - The Categorization maps DataStructureDefintions to Categories, when
    #   there could be many.
    # list(cat_response.category_scheme['MOBILE_NAVI']['07'])

    dfs = sdmx.to_pandas(msg1.dataflow).head()
    expected = pd.Series({
        'AME': 'AMECO',
        'BKN': 'Banknotes statistics',
        'BLS': 'Bank Lending Survey Statistics',
        'BOP': ('Euro Area Balance of Payments and International Investment '
                'Position Statistics'),
        'BSI': 'Balance Sheet Items',
        })
    assert_pd_equal(dfs, expected)

    flows = ecb.dataflow()  # noqa: F841
    dsd_id = msg1.dataflow.EXR.structure.id
    assert dsd_id == 'ECB_EXR1'

    refs = dict(references='all')
    msg2 = ecb.datastructure(resource_id=dsd_id, params=refs)
    dsd = msg2.structure[dsd_id]

    assert sdmx.to_pandas(dsd.dimensions) == [
        'FREQ', 'CURRENCY', 'CURRENCY_DENOM', 'EXR_TYPE', 'EXR_SUFFIX',
        'TIME_PERIOD']

    cl = sdmx.to_pandas(msg2.codelist['CL_CURRENCY']).sort_index()
    expected = pd.Series({
        'ADF': 'Andorran Franc (1-1 peg to the French franc)',
        'ADP': 'Andorran Peseta (1-1 peg to the Spanish peseta)',
        'AED': 'United Arab Emirates dirham',
        'AFA': 'Afghanistan afghani (old)',
        'AFN': 'Afghanistan, Afghanis',
        }, name='Currency code list') \
        .rename_axis('CL_CURRENCY')
    assert_pd_equal(cl.head(), expected)


@pytest.mark.remote_data
def test_doc_usage_data():
    """Code examples in usage.rst."""
    ecb = Request('ECB')

    data_response = ecb.data(resource_id='EXR', key={'CURRENCY': 'USD+JPY'},
                             params={'startPeriod': '2016',
                                     'endPeriod': '2016-12-31'})
    # # Commented: do the same without triggering requests for validation
    # data_response = ecb.data(resource_id='EXR', key='.JPY+USD...',
    #                          params={'startPeriod': '2016',
    #                                  'endPeriod': '2016-12-31'})
    data = data_response.data[0]

    assert type(data) is GenericDataSet

    # This message doesn't explicitly specify the remaining dimensions; unless
    # they are inferred from the SeriesKeys, then the DimensionDescriptor is
    # not complete
    # assert data.structured_by.dimensions[-1] == 'TIME_PERIOD'
    # data.dim_at_obs

    series_keys = list(data.series)

    assert len(series_keys) == 16

    series_keys[5]

    assert (sorted(set(sk.FREQ.value for sk in data.series))
            == 'A D H M Q'.split())

    daily = sdmx.to_pandas(data).xs('D', level='FREQ')
    assert len(daily) == 514

    assert_pd_equal(daily.tail().values,
                    np.array([1.0446, 1.0445, 1.0401, 1.0453, 1.0541]))


def test_doc_howto_timeseries():
    with specimen('sg-ts.xml') as f:
        ds = sdmx.read_sdmx(f).data[0]

    # Convert to pd.Series and unstack the time dimension to columns
    base = sdmx.to_pandas(ds)
    s1 = base.unstack('TIME_PERIOD')

    # DatetimeIndex on columns
    s1.columns = pd.to_datetime(s1.columns)
    assert isinstance(s1.columns, pd.DatetimeIndex)

    # DatetimeIndex on index
    s2 = base.unstack('TIME_PERIOD').transpose()
    s2.index = pd.to_datetime(s2.index)
    assert isinstance(s2.index, pd.DatetimeIndex)

    # Same with pd.PeriodIndex
    s3 = s1.to_period(axis=1)
    assert isinstance(s3.columns, pd.PeriodIndex)
    assert s3.columns.freqstr == 'M'

    s4 = s2.to_period(axis=0)
    assert isinstance(s4.index, pd.PeriodIndex)
    assert s4.index.freqstr == 'M'



import pandasdmx as sdmx

from .data import specimen


EXPECTED = {
    # Structure messages
    'IPI-2010-A21-structure.xml': """<pandasdmx.StructureMessage>
  <Header>
    id: 'categorisation_1450864290565'
    prepared: '2015-12-23T09:51:30.565Z'
    receiver: 'unknown'
    sender: 'FR1'
  CategoryScheme (1): CLASSEMENT_DATAFLOWS
  Codelist (3): CL_FREQ CL_NAF2_A21 CL_NATURE
  ConceptScheme (1): CONCEPTS_INSEE
  DataflowDefinition (1): IPI-2010-A21
  DataStructureDefinition (1): IPI-2010-A21""",

    # This message shows the summarization feature: the DFD list is truncated
    'dataflow.xml': """<pandasdmx.StructureMessage>
  <Header>
    id: 'dataflow_ENQ-CONJ-TRES-IND-PERSP_1450865196042'
    prepared: '2015-12-23T10:06:36.042Z'
    receiver: 'unknown'
    sender: 'FR1'
  DataflowDefinition (663): ACT-TRIM-ANC BPM6-CCAPITAL BPM6-CFINANCIER ...""",

    # Data message
    'sg-xs.xml': """<pandasdmx.DataMessage>
  <Header>
    id: 'Generic'
    prepared: '2010-01-04T16:21:49+01:00'
    sender: 'ECB'
  DataSet (1)
  dataflow: <DataflowDefinition: 'STR1'=''>
  observation_dimension: <Dimension: CURRENCY>""",

    # This message has two DataSets:
    'action-delete.json': """<pandasdmx.DataMessage>
  <Header>
    id: '62b5f19d-f1c9-495d-8446-a3661ed24753'
    prepared: '2012-11-29T08:40:26Z'
    sender: <Item: 'ECB'='European Central Bank'>
  DataSet (2)
  dataflow: <DataflowDefinition: 'None'=''>
  observation_dimension: [<Dimension: CURRENCY>]""",
}


def test_message_repr():
    for pattern, expected in EXPECTED.items():
        with specimen(pattern) as f:
            msg = sdmx.read_sdmx(f)
        assert expected == repr(msg)



# TODO tidy these tests to use fixtures/methods from pandasdmx.tests
from collections import OrderedDict

import pandasdmx as sdmx
from pandasdmx import Request
import pytest

from .data import BASE_PATH as test_data_path


test_data_path = test_data_path / 'INSEE'

DATAFLOW_FP = test_data_path / 'dataflow.xml'

DATASETS = {
    'IPI-2010-A21': {
        'data-fp': test_data_path / 'IPI-2010-A21.xml',
        'datastructure-fp': test_data_path / 'IPI-2010-A21-structure.xml',
        'series_count': 20,
    },
    'CNA-2010-CONSO-SI-A17': {
        'data-fp': test_data_path / 'CNA-2010-CONSO-SI-A17.xml',
        'datastructure-fp': (test_data_path /
                             'CNA-2010-CONSO-SI-A17-structure.xml'),
        'series_count': 1,
    },
}

SERIES = {
    'UNEMPLOYMENT_CAT_A_B_C': {
        'data-fp': test_data_path / 'bug-series-freq-data.xml',
    }
}


class TestINSEE:
    @pytest.fixture(scope='class')
    def req(self):
        return Request('INSEE')

    def test_load_dataset(self, req):
        dataset_code = 'IPI-2010-A21'

        # Load all dataflows
        dataflows_response = sdmx.read_sdmx(DATAFLOW_FP)
        dataflows = dataflows_response.dataflow

        assert len(dataflows) == 663
        assert dataset_code in dataflows

        # Load datastructure for current dataset_code
        fp_datastructure = DATASETS[dataset_code]['datastructure-fp']
        datastructure_response = sdmx.read_sdmx(fp_datastructure)
        assert dataset_code in datastructure_response.dataflow
        dsd = datastructure_response.dataflow[dataset_code].structure

        # Verify dimensions list
        dimensions = OrderedDict([dim.id, dim] for dim in
                                 dsd.dimensions if dim.id not in
                                 ['TIME', 'TIME_PERIOD'])
        dim_keys = list(dimensions.keys())
        assert dim_keys == ['FREQ', 'PRODUIT', 'NATURE']

        # Load datas for the current dataset
        fp_data = DATASETS[dataset_code]['data-fp']
        data = sdmx.read_sdmx(fp_data)

        # Verify series count and values
        series = data.data[0].series
        series_count = len(series)
        assert series_count == DATASETS[dataset_code]['series_count']

        first_series = series[0]
        observations = first_series

        first_obs = observations[0]
        last_obs = observations[-1]

        assert first_obs.dim == '2015-10'
        assert first_obs.value == '105.61'

        assert last_obs.dim == '1990-01'
        assert last_obs.value == '139.22'

    def test_fixe_key_names(self, req):
        """Verify key or attribute contains '-' in name."""
        dataset_code = 'CNA-2010-CONSO-SI-A17'

        fp_datastructure = DATASETS[dataset_code]['datastructure-fp']
        datastructure_response = sdmx.read_sdmx(fp_datastructure)
        assert dataset_code in datastructure_response.dataflow
        dsd = datastructure_response.dataflow[dataset_code].structure

        dimensions = OrderedDict([dim.id, dim] for dim in
                                 dsd.dimensions if dim.id not in
                                 ['TIME', 'TIME_PERIOD'])
        dim_keys = list(dimensions.keys())
        assert dim_keys == ['SECT-INST', 'OPERATION', 'PRODUIT', 'PRIX']

        fp_data = DATASETS[dataset_code]['data-fp']
        data = sdmx.read_sdmx(fp_data)
        series = data.data[0].series
        series_key = list(series.keys())[0]

        assert (list(series_key.values.keys()) ==
                ['SECT-INST', 'OPERATION', 'PRODUIT', 'PRIX'])

        assert (list(series_key.attrib.keys()) ==
                ['FREQ', 'IDBANK', 'TITLE', 'LAST_UPDATE', 'UNIT_MEASURE',
                 'UNIT_MULT', 'REF_AREA', 'DECIMALS', 'BASE_PER',
                 'TIME_PER_COLLECT'])

    def test_freq_in_series_attribute(self, req):
        # Test that we don't have regression on Issues #39 and #41
        # INSEE time series provide the FREQ value as attribute on the series
        # instead of a dimension. This caused a runtime error when writing as
        # pandas dataframe.
        data_response = sdmx.read_sdmx(
            SERIES['UNEMPLOYMENT_CAT_A_B_C']['data-fp'])
        sdmx.to_pandas(data_response)



import pandas as pd

import pandasdmx as sdmx

from .data import specimen


def pd_obj(name):
    """Return the specimen at *name* read, then converted to pandas."""
    with specimen(name) as f:
        data_msg = sdmx.read_sdmx(f)
    return sdmx.to_pandas(data_msg, rtype='compat')


def test_xml():
    """Test that objects have the expected layout with rtype='compat'."""
    # GenericData with non-time Dimension at observation → data frame
    df1 = pd_obj('ng-xs.xml')
    assert isinstance(df1.index, pd.Index) and df1.index.name == 'CURRENCY'
    assert isinstance(df1.columns, pd.MultiIndex)

    # TimeSeriesData → data frame with DatetimeIndex
    df2 = pd_obj('ng-ts.xml')
    assert isinstance(df2.index, pd.DatetimeIndex)
    assert isinstance(df2.columns, pd.MultiIndex)

    # GenericData with AllDimensions at observation → series
    df3 = pd_obj('ng-flat.xml')
    assert isinstance(df3, pd.Series)
    assert isinstance(df3.index, pd.MultiIndex)



import pandasdmx as sdmx
import pytest

from .data import specimen, test_files


@pytest.mark.parametrize('path', **test_files(format='json'))
def test_json_read(path):
    """Test that the samples from the SDMX-JSON spec can be read."""
    sdmx.read_sdmx(path)


def test_header():
    with specimen('flat.json') as f:
        resp = sdmx.read_sdmx(f)
    assert resp.header.id == '62b5f19d-f1c9-495d-8446-a3661ed24753'



from requests import HTTPError  # noqa: F401


class ParseError(Exception):
    """:class:`pandasdmx.reader.Reader` is unable to parse a message."""
    pass


class XMLParseError(ParseError):
    """Error in parsing SDMX-ML.

    Parameters
    ----------
    reader : pandasdmx.reader.sdmxml.Reader
    elem : lxml.etree.Element
    """
    def __init__(self, reader, elem, message=None):
        self.stack = reader._stack
        self.elem = elem
        self.message = message
        self.__suppress_context__ = True

    def __str__(self):
        from lxml import etree

        msg = '\n\n'.join(filter(None, [
            self.message,
            '{}: {}'.format(self.__cause__.__class__.__name__,
                            self.__cause__),
            'Stack:\n' + '\n> '.join(self.stack),
            etree.tostring(self.elem, pretty_print=True).decode(),
            ]))
        self.__cause__ = None
        return msg



from . import Source as BaseSource


class Source(BaseSource):
    _id = 'WB'

    def modify_request_args(self, kwargs):
        """World Bank's agency ID."""
        super().modify_request_args(kwargs)

        kwargs.setdefault('provider', 'WBG_WITS')



from itertools import chain

import numpy as np
import pandas as pd

from pandasdmx import model
from pandasdmx.model import (
    DEFAULT_LOCALE,
    AgencyScheme,
    AllDimensions,
    DataAttribute,
    DataflowDefinition,
    DataStructureDefinition,
    DataSet,
    Dimension,
    DimensionComponent,
    # DimensionDescriptor,
    CategoryScheme,
    Codelist,
    Component,
    ConceptScheme,
    ItemScheme,
    NameableArtefact,
    Observation,
    SeriesKey,
    TimeDimension,
)
from pandasdmx.util import DictLike


#: Default return type for :func:`write_dataset` and similar methods. Either
#: 'compat' or 'rows'. See the ref:`HOWTO <howto-rtype>`.
DEFAULT_RTYPE = 'compat'


# Class → common write_*() methods
_ALIAS = {
    DictLike: dict,
    AgencyScheme: ItemScheme,
    CategoryScheme: ItemScheme,
    ConceptScheme: ItemScheme,
    Codelist: ItemScheme,
    DataflowDefinition: NameableArtefact,
    DataStructureDefinition: NameableArtefact,
    Dimension: Component,
    TimeDimension: Component,
    model.GenericDataSet: DataSet,
    model.GenericTimeSeriesDataSet: DataSet,
    model.StructureSpecificDataSet: DataSet,
    model.StructureSpecificTimeSeriesDataSet: DataSet,
}


def write(obj, *args, **kwargs):
    """Convert an SDMX *obj* to :mod:`pandas` object(s).

    Implements a dispatch pattern according to the type of *obj*. For instance,
    a :class:`.DataSet` object is converted using :func:`.write_dataset`. See
    individual ``write_*`` methods named for more information on their
    behaviour, including accepted *args* and *kwargs*.
    """
    cls = obj.__class__
    function = 'write_' + _ALIAS.get(cls, cls).__name__.lower()
    return globals()[function](obj, *args, **kwargs)


# Functions for Python containers
def write_list(obj, *args, **kwargs):
    """Convert a :class:`list` of SDMX objects.

    For the following *obj*, :meth:`write_list` returns :class:`pandas.Series`
    instead of a :class:`list`:

    - a list of :class:`Observation <pandasdmx.model.Observation>`:
      the Observations are written using :meth:`write_dataset`.
    - a list with only 1 :class:`DataSet <pandasdmx.model.DataSet>` (e.g. the
       :attr:`data <pandasdmx.message.DataMessage.data>` attribute of
       :class:`DataMessage <pandasdmx.message.DataMessage>`): the Series for
       the single element is returned.
    - a list of :class:`SeriesKey`: the key values (but no data) are returned.
    """
    if isinstance(obj[0], Observation):
        return write_dataset(obj, *args, **kwargs)
    elif isinstance(obj[0], DataSet) and len(obj) == 1:
        return write(obj[0], *args, **kwargs)
    elif isinstance(obj[0], SeriesKey):
        return write_serieskeys(obj, *args, **kwargs)
    else:
        return [write(item, *args, **kwargs) for item in obj]


def write_dict(obj, *args, **kwargs):
    """Convert mappings.

    The values of the mapping are write()'d individually. If the resulting
    values are :class:`str` or :class:`pd.Series` *with indexes that share the
    same name*, then they are converted to a pd.Series, possibly with a
    pd.MultiIndex. Otherwise, a DictLike is returned.
    """
    result = {k: write(v, *args, **kwargs) for k, v in obj.items()}

    result_type = set(type(v) for v in result.values())

    if result_type <= {pd.Series, pd.DataFrame}:
        if (len(set(map(lambda s: s.index.name, result.values()))) == 1 and
                len(result) > 1):
            # Can safely concatenate these to a pd.MultiIndex'd Series.
            return pd.concat(result)
        else:
            # The individual pd.Series are indexed by different dimensions; do
            # not concatenate.
            return DictLike(result)
    elif result_type == {str}:
        return pd.Series(result)
    elif result_type == {DictLike}:
        return result
    elif result_type == set():
        # No results
        return pd.Series()
    else:
        raise ValueError(result_type)


def write_set(obj, *args, **kwargs):
    """Convert :class:`set`."""
    result = {write(o, *args, **kwargs) for o in obj}
    return result


# Functions for message classes
def write_datamessage(obj, *args, rtype=None, **kwargs):
    """Convert :class:`.DataMessage`.

    The data set(s) within the message are converted to pandas objects.

    Parameters
    ----------
    rtype : 'compat' or 'rows', optional
        Data type to return; default :data:`.DEFAULT_RTYPE`. See the
        :ref:`HOWTO <howto-rtype>`.
    kwargs :
        Passed to :meth:`write_dataset` for each data set.

    Returns
    -------
    :class:`pandas.Series` or :class:`pandas.DataFrame`
        if `obj` has only one data set.
    list of (:class:`pandas.Series` or :class:`pandas.DataFrame`)
        if `obj` has more than one data set.
    """
    # Pass the message's DSD to assist datetime handling
    kwargs.setdefault('dsd', obj.dataflow.structure)

    # Pass the return type and associated information
    kwargs['_rtype'] = rtype or DEFAULT_RTYPE
    if kwargs['_rtype'] == 'compat':
        kwargs['_message_class'] = obj.__class__
        kwargs['_observation_dimension'] = obj.observation_dimension

    if len(obj.data) == 1:
        return write(obj.data[0], *args, **kwargs)
    else:
        return [write(ds, *args, **kwargs) for ds in obj.data]


def write_structuremessage(obj, include=None, **kwargs):
    """Convert :class:`.StructureMessage`.

    Parameters
    ----------
    obj : .StructureMessage
    include : iterable of str or str, optional
        One or more of the attributes of the StructureMessage (
        'category_scheme', 'codelist', etc.) to transform.
    kwargs :
        Passed to :meth:`write` for each attribute.

    Returns
    -------
    .DictLike
        Keys are StructureMessage attributes; values are pandas objects.
    """
    all_contents = {
        'category_scheme',
        'codelist',
        'concept_scheme',
        'constraint',
        'dataflow',
        'structure',
        'organisation_scheme',
    }

    # Handle arguments
    if include is None:
        attrs = all_contents
    else:
        attrs = set([include] if isinstance(include, str) else include)
        # Silently discard invalid names
        attrs &= all_contents
    attrs = sorted(attrs)

    result = DictLike()
    for a in attrs:
        dl = write(getattr(obj, a), **kwargs)
        if len(dl):
            # Only add non-empty elements
            result[a] = dl

    return result


# Functions for model classes

def write_component(obj):
    """Convert :class:`.Component`.

    The :attr:`~.Concept.id` attribute of the
    :attr:`~.Component.concept_identity` is returned.
    """
    return str(obj.concept_identity.id)


def write_contentconstraint(obj, **kwargs):
    """Convert :class:`.ContentConstraint`."""
    if len(obj.data_content_region) != 1:
        raise NotImplementedError

    return write(obj.data_content_region[0], **kwargs)


def write_cuberegion(obj, **kwargs):
    """Convert :class:`.CubeRegion`."""
    result = DictLike()
    for dim, memberselection in obj.member.items():
        result[dim] = pd.Series([mv.value for mv in memberselection.values],
                                name=dim.id)
    return result


def write_dataset(obj, attributes='', dtype=np.float64, constraint=None,
                  datetime=False, **kwargs):
    """Convert :class:`~.DataSet`.

    See the :ref:`walkthrough <datetime>` for examples of using the `datetime`
    argument.

    Parameters
    ----------
    obj : :class:`~.DataSet` or iterable of :class:`~.Observation`
    attributes : str
        Types of attributes to return with the data. A string containing
        zero or more of:

        - ``'o'``: attributes attached to each :class:`~.Observation` .
        - ``'s'``: attributes attached to any (0 or 1) :class:`~.SeriesKey`
          associated with each Observation.
        - ``'g'``: attributes attached to any (0 or more) :class:`~.GroupKey`
          associated with each Observation.
        - ``'d'``: attributes attached to the :class:`~.DataSet` containing the
          Observations.

    dtype : str or :class:`numpy.dtype` or None
        Datatype for values. If None, do not return the values of a series.
        In this case, `attributes` must not be an empty string so that some
        attribute is returned.
    constraint : .ContentConstraint, optional
        If given, only Observations included by the *constraint* are returned.
    datetime : bool or str or or .Dimension or dict, optional
        If given, return a DataFrame with a :class:`~pandas.DatetimeIndex`
        or :class:`~pandas.PeriodIndex` as the index and all other dimensions
        as columns. Valid `datetime` values include:

        - :class:`bool`: if :obj:`True`, determine the time dimension
          automatically by detecting a :class:`~.TimeDimension`.
        - :class:`str`: ID of the time dimension.
        - :class:`~.Dimension`: the matching Dimension is the time dimension.
        - :class:`dict`: advanced behaviour. Keys may include:

          - **dim** (:class:`~.Dimension` or :class:`str`): the time dimension
            or its ID.
          - **axis** (`{0 or 'index', 1 or 'columns'}`): axis on which to place
            the time dimension (default: 0).
          - **freq** (:obj:`True` or :class:`str` or :class:`~.Dimension`):
            produce :class:`pandas.PeriodIndex`. If :class:`str`, the ID of a
            Dimension containing a frequency specification. If a Dimension, the
            specified dimension is used for the frequency specification.

            Any Dimension used for the frequency specification is does not
            appear in the returned DataFrame.

    Returns
    -------
    :class:`pandas.DataFrame`
        - if `attributes` is not ``''``, a data frame with one row per
          Observation, ``value`` as the first column, and additional columns
          for each attribute;
        - if `datetime` is given, various layouts as described above; or
        - if `_rtype` (passed from :func:`write_datamessage`) is 'compat',
          various layouts as described in the :ref:`HOWTO <howto-rtype>`.
    :class:`pandas.Series` with :class:`pandas.MultiIndex`
        Otherwise.
    """
    # If called directly on a DataSet (rather than a parent DataMessage),
    # cannot determine the "dimension at observation level"
    rtype = kwargs.setdefault('_rtype', 'rows')

    # Validate attributes argument
    attributes = attributes or ''
    try:
        attributes = attributes.lower()
    except AttributeError:
        raise TypeError("'attributes' argument must be str")

    if rtype == 'compat' and \
            kwargs['_observation_dimension'] is not AllDimensions:
        # Cannot return attributes in this case
        attributes = ''
    elif set(attributes) - {'o', 's', 'g', 'd'}:
        raise ValueError(f"attributes must be in 'osgd'; got {attributes}")

    # Iterate on observations
    result = {}
    for observation in getattr(obj, 'obs', obj):
        # Check that the Observation is within the constraint, if any
        key = observation.key.order()
        if constraint and key not in constraint:
            continue

        # Add value and attributes
        row = {}
        if dtype:
            row['value'] = observation.value
        if attributes:
            row.update(observation.attrib)

        result[tuple(map(str, key.get_values()))] = row

    result = pd.DataFrame.from_dict(result, orient='index')

    if len(result):
        result.index.names = observation.key.order().values.keys()
        if dtype:
            result['value'] = result['value'].astype(dtype)
            if not attributes:
                result = result['value']

    # Reshape for compatibility with v0.9
    result, datetime, kwargs = _dataset_compat(result, datetime, kwargs)
    # Handle the datetime argument, if any
    return _maybe_convert_datetime(result, datetime, obj=obj, **kwargs)


def _dataset_compat(df, datetime, kwargs):
    """Helper for :meth:`.write_dataset` 0.9 compatibility."""
    rtype = kwargs.pop('_rtype')
    if rtype != 'compat':
        return df, datetime, kwargs  # Do nothing

    # Remove compatibility arguments from kwargs
    kwargs.pop('_message_class')
    obs_dim = kwargs.pop('_observation_dimension')

    if isinstance(obs_dim, list) and len(obs_dim) == 1:
        # Unwrap a length-1 list
        obs_dim = obs_dim[0]

    if obs_dim in (AllDimensions, None):
        pass  # Do nothing
    elif isinstance(obs_dim, TimeDimension):
        # Don't modify *df*; only change arguments so that
        # _maybe_convert_datetime performs the desired changes
        if datetime is False or datetime is True:
            # Either datetime is not given, or True without specifying a
            # dimension; overwrite
            datetime = obs_dim
        elif isinstance(datetime, dict):
            # Dict argument; ensure the 'dim' key is the same as obs_dim
            if datetime.setdefault('dim', obs_dim) != obs_dim:
                msg = (f"datetime={datetime} conflicts with rtype='compat' and"
                       f" {obs_dim} at observation level")
                raise ValueError(msg)
        else:
            assert datetime == obs_dim, (datetime, obs_dim)
    elif isinstance(obs_dim, DimensionComponent):
        # Pivot all levels except the observation dimension
        df = df.unstack([n for n in df.index.names if n != obs_dim.id])
    else:
        # E.g. some JSON messages have two dimensions at the observation level;
        # behaviour is unspecified here, so do nothing.
        pass

    return df, datetime, kwargs


def _maybe_convert_datetime(df, arg, obj, dsd=None):
    """Helper for :meth:`.write_dataset` to handle datetime indices.

    Parameters
    ----------
    df : pandas.DataFrame
    arg : dict
        From the `datetime` argument to :meth:`write_dataset`.
    obj :
        From the `obj` argument to :meth:`write_dataset`.
    dsd: ~.DataStructureDefinition, optional
    """
    if not arg:
        # False, None, empty dict: no datetime conversion
        return df

    # Check argument values
    param = dict(dim=None, axis=0, freq=False)
    if isinstance(arg, str):
        param['dim'] = arg
    elif isinstance(arg, DimensionComponent):
        param['dim'] = arg.id
    elif isinstance(arg, dict):
        extra_keys = set(arg.keys()) - set(param.keys())
        if extra_keys:
            raise ValueError(extra_keys)
        param.update(arg)
    elif isinstance(arg, bool):
        pass  # True
    else:
        raise ValueError(arg)

    def _get_dims():
        """Return an appropriate list of dimensions."""
        if len(obj.structured_by.dimensions.components):
            return obj.structured_by.dimensions.components
        elif dsd:
            return dsd.dimensions.components
        else:
            return []

    def _get_attrs():
        """Return an appropriate list of attributes."""
        if len(obj.structured_by.attributes.components):
            return obj.structured_by.attributes.components
        elif dsd:
            return dsd.attributes.components
        else:
            return []

    if not param['dim']:
        # Determine time dimension
        dims = _get_dims()
        for dim in dims:
            if isinstance(dim, TimeDimension):
                param['dim'] = dim
                break
        if not param['dim']:
            raise ValueError(f'no TimeDimension in {dims}')

    # Unstack all but the time dimension and convert
    other_dims = list(filter(lambda d: d != param['dim'], df.index.names))
    df = df.unstack(other_dims)
    df.index = pd.to_datetime(df.index)

    if param['freq']:
        # Determine frequency string, Dimension, or Attribute
        freq = param['freq']
        if isinstance(freq, str) and freq not in pd.offsets.prefix_mapping:
            # ID of a Dimension or Attribute
            for component in chain(_get_dims(), _get_attrs()):
                if component.id == freq:
                    freq = component
                    break

            # No named dimension in the DSD; but perhaps on the df
            if isinstance(freq, str):
                if freq in df.columns.names:
                    freq = Dimension(id=freq)
                else:
                    raise ValueError(freq)

        if isinstance(freq, Dimension):
            # Retrieve Dimension values from pd.MultiIndex level
            level = freq.id
            i = df.columns.names.index(level)
            values = set(df.columns.levels[i])

            if len(values) > 1:
                values = sorted(values)
                raise ValueError('cannot convert to PeriodIndex with '
                                 f'non-unique freq={values}')

            # Store the unique value
            freq = values.pop()

            # Remove the index level
            df.columns = df.columns.droplevel(i)
        elif isinstance(freq, DataAttribute):  # pragma: no cover
            raise NotImplementedError

        df.index = df.index.to_period(freq=freq)

    if param['axis'] in {1, 'columns'}:
        # Change axis
        df = df.transpose()

    return df


def write_dimensiondescriptor(obj):
    """Convert :class:`.DimensionDescriptor`.

    The :attr:`~.DimensionDescriptor.components` of the DimensionDescriptor
    are written.
    """
    return write(obj.components)


def write_itemscheme(obj, locale=DEFAULT_LOCALE):
    """Convert :class:`.ItemScheme`.

    Names from *locale* are serialized.

    Returns
    -------
    pandas.Series
    """
    items = {}
    seen = set()

    def add_item(item):
        """Recursive helper for adding items."""
        # Track seen items
        if item in seen:
            return
        else:
            seen.add(item)

        # Localized name
        row = {'name': item.name.localized_default(locale)}
        try:
            # Parent ID
            row['parent'] = item.parent.id
        except AttributeError:
            row['parent'] = ''

        items[item.id] = row

        # Add this item's children, recursively
        for child in item.child:
            add_item(child)

    for item in obj:
        add_item(item)

    # Convert to DataFrame
    result = pd.DataFrame.from_dict(items, orient='index', dtype=object) \
               .rename_axis(obj.id, axis='index')

    if len(result) and not result['parent'].str.len().any():
        # 'parent' column is empty; convert to pd.Series and rename
        result = result['name'].rename(obj.name.localized_default(locale))

    return result


def write_membervalue(obj):
    """Convert :class:`.MemberValue`."""
    return obj.value


def write_nameableartefact(obj):
    """Convert :class:`.NameableArtefact`.

    The :attr:`~.NameableArtefact.name` attribute of *obj* is returned.
    """
    return str(obj.name)


def write_serieskeys(obj):
    """Convert a list of :class:`.SeriesKey`."""
    result = []
    for sk in obj:
        result.append({dim: kv.value for dim, kv in sk.order().values.items()})
    # TODO perhaps return as a pd.MultiIndex if that is more useful
    return pd.DataFrame(result)



from abc import ABC, abstractmethod


class BaseReader(ABC):
    @abstractmethod
    def read_message(self, source):
        """Read message from *source*.

        Must return an instance of a model.Message subclass.
        """
        pass  # pragma: no cover

    # Backwards-compatibility
    def initialize(self, source):
        return self.read_message(source)


# TODO make an attribute of Reader
CONTENT_TYPES = {
    'xml': [
        'application/xml',
        'application/vnd.sdmx.genericdata+xml',
        'application/vnd.sdmx.structure+xml',
        'application/vnd.sdmx.structurespecificdata+xml',
        'text/xml',
    ],
    'json': [
        'text/json',
        # For e.g. OECD
        'draft-sdmx-json',
        'application/vnd.sdmx.draft-sdmx-json+json',
    ],
}


def get_reader_for_content_type(ctype):
    ctype = str(ctype).split(';')[0].strip()
    if ctype in CONTENT_TYPES['xml']:
        from .sdmxml import Reader
        return Reader
    elif ctype in CONTENT_TYPES['json']:
        from .sdmxjson import Reader
        return Reader
    else:
        raise ValueError(ctype)



import importlib
from distutils import version

import numpy as np
import pandas as pd
import pytest

from .data import BASE_PATH


def assert_pd_equal(left, right, **kwargs):
    """Assert equality of two pandas objects."""
    if left is None:
        return
    method = {
        pd.Series: pd.testing.assert_series_equal,
        pd.DataFrame: pd.testing.assert_frame_equal,
        np.ndarray: np.testing.assert_array_equal,
        }[left.__class__]
    method(left, right, **kwargs)


class MessageTest:
    path = BASE_PATH
    filename = None

    @pytest.fixture(scope='class')
    def msg(self):
        import pandasdmx as sdmx
        return sdmx.read_sdmx(self.path / self.filename)


# thanks to xarray
def _importorskip(modname, minversion=None):
    try:
        mod = importlib.import_module(modname)
        has = True
        if minversion is not None:
            if LooseVersion(mod.__version__) < LooseVersion(minversion):
                raise ImportError('Minimum version not satisfied')
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason='requires {}'.format(modname))
    return has, func


def LooseVersion(vstring):
    # When the development version is something like '0.10.9+aac7bfc', this
    # function will just discard the git commit id.
    vstring = vstring.split('+')[0]
    return version.LooseVersion(vstring)


has_requests_cache, requires_requests_cache = _importorskip('requests_cache')



from lxml.etree import Element
import pandasdmx as sdmx
from pandasdmx.model import (
    Facet, FacetType, FacetValueType,
    )
from pandasdmx.reader.sdmxml import XMLParseError, Reader
import pytest

from .data import specimen, test_files


# Read example data files
@pytest.mark.parametrize('path', **test_files(format='xml', kind='data'))
def test_read_xml(path):
    sdmx.read_sdmx(path)


# Read example structure files
@pytest.mark.parametrize('path', **test_files(format='xml', kind='structure'))
def test_read_xml_structure(path):
    sdmx.read_sdmx(path)


def test_read_xml_structure_insee():
    with specimen('IPI-2010-A21-structure.xml') as f:
        msg = sdmx.read_sdmx(f)

    # Same objects referenced
    assert (id(msg.dataflow['IPI-2010-A21'].structure) ==
            id(msg.structure['IPI-2010-A21']))

    # Number of dimensions loaded correctly
    dsd = msg.structure['IPI-2010-A21']
    assert len(dsd.dimensions) == 4


# Read structure-specific messages
def test_read_ss_xml():
    with specimen('M.USD.EUR.SP00.A.xml', opened=False) as f:
        msg_path = f
        dsd_path = f.parent / 'structure.xml'

    # Read the DSD
    dsd = sdmx.read_sdmx(dsd_path).structure['ECB_EXR1']

    # Read a data message
    msg = sdmx.read_sdmx(msg_path, dsd=dsd)
    ds = msg.data[0]

    # The dataset in the message is structured by the DSD
    assert ds.structured_by is dsd

    # Structures referenced in the dataset are from the dsd

    s0_key = list(ds.series.keys())[0]

    # AttributeValue.value_for
    assert s0_key.attrib['DECIMALS'].value_for \
        is dsd.attributes.get('DECIMALS')

    # SeriesKey.described_by
    assert s0_key.described_by is dsd.dimensions

    # Key.described_by
    assert ds.obs[0].key.described_by is dsd.dimensions

    # KeyValue.value_for
    assert ds.obs[0].key.values[0].value_for \
        is dsd.dimensions.get('FREQ')

    # DSD information that is not in the data message can be looked up through
    # navigating object relationships
    TIME_FORMAT = s0_key.attrib['TIME_FORMAT'].value_for
    assert len(TIME_FORMAT.related_to.dimensions) == 5


E = Element

# Each entry is a tuple with 2 elements:
# 1. an instance of lxml.etree.Element to be parsed.
# 2. Either:
#   - A pandasdmx.model object, in which case the parsed element must match
#     the object.
#   - A string, in which case parsing the element is expected to fail, raising
#     an exception matching the string.
ELEMENTS = [
    # Reader.parse_facet
    (E('Facet', isSequence='False', startValue='3.4', endValue='1'), None),
    # …attribute names are munged; default textType is supplied
    (E('EnumerationFormat', minLength='1', maxLength='6'),
     Facet(type=FacetType(min_length=1, max_length=6),
           value_type=FacetValueType['string'])),
    # …invalid attributes cause an exception
    (E('TextFormat', invalidFacetTypeAttr='foo'),
     'ValueError: "FacetType" object has no field "invalid_facet_type_attr"'),
]


@pytest.mark.parametrize('elem, expected', ELEMENTS,
                         ids=list(map(str, range(len(ELEMENTS)))))
def test_parse_elem(elem, expected):
    """Test individual XML elements.

    This method allows unit-level testing of specific XML elements appearing in
    SDMX-ML messages. Add elements by extending the list passed to the
    parametrize() decorator.
    """
    # _parse() operates on the child elements of the first argument. Create a
    # temporary element to contain *elem*
    tmp = Element('root')
    tmp.append(elem)

    # Create a reader
    reader = Reader()

    if isinstance(expected, str):
        # Parsing the element raises an exception
        with pytest.raises(XMLParseError, match=expected):
            reader._parse(tmp)
    else:
        # The element is parsed successfully
        result = reader._parse(tmp).popitem()[1]

        if expected:
            # Expected value supplied
            assert expected == result



'''Statistical Data and Metadata eXchange (SDMX) for the Python data ecosystem'''

from pandasdmx.api import Request, read_sdmx, read_url
from pandasdmx.source import add_source, list_sources
from pandasdmx.util import Resource
from pandasdmx.writer import write as to_pandas
import logging

__all__ = [
    'Request',
    'Resource',
    'list_sources',
    'read_sdmx',
    'read_url'
    ]


__version__ = '1.0.1'


#: Top-level logger for pandaSDMX.
logger = logging.getLogger(__name__)


def _init_logger():
    handler = logging.StreamHandler()
    fmt = '{asctime} {name} - {levelname}: {message}'
    handler.setFormatter(logging.Formatter(fmt, style='{'))
    logger.addHandler(handler)
    logger.setLevel(logging.ERROR)


_init_logger()



"""Network requests API.

This module defines :class:`.Request`, which forms the high-level API of
:mod:`pandasdmx`. Requesting data and metadata from an SDMX server requires a
understanding of this API and a basic understanding of the SDMX web service
guidelines.
"""
from functools import partial
import logging
from pathlib import Path
from warnings import warn

from pandasdmx import remote
from .model import DataStructureDefinition, MaintainableArtefact
from .reader import get_reader_for_content_type
from .source import NoSource, list_sources, sources
from .util import Resource
import requests

logger = logging.getLogger(__name__)


class Request:
    """Client for a SDMX REST web service.

    Parameters
    ----------
    source : str or :class:`pandasdmx.source.Source`
        Identifier of a data source. If a string, must be one of the known
        sources in :meth:`list_sources`.
    log_level : int
        Override the package-wide logger with one of the
        :ref:`standard logging levels <py:levels>`.
    session : optional instance of :class:`requests.Session`, typically  a subclass. If given,
        it is  used for HTTP requests, and any   *session_opts* passed  will raise TypeError. 
        One  use case is the injection of alternative caching libraries such as Cache Control.     
    **session_opts
        Additional keyword arguments are passed to
        :class:`pandasdmx.remote.Session`.

    """
    cache = {}

    #: :class:`.source.Source` for requests sent from the instance.
    source = None

    #: :class:`.Session` for queries sent from the instance.
    session = None

    def __init__(self, source=None, log_level=None, 
            session=None, **session_opts):
        """Constructor."""
        try:
            self.source = sources[source.upper()] if source else NoSource
        except KeyError:
            raise ValueError('source must be None or one of: %s' %
                             ' '.join(list_sources()))

        if session and session_opts:
            raise TypeError('When `session` is given, `session_opts` must be  empty.')
        self.session = (session or 
            remote.Session(**session_opts))

        if log_level:
            logging.getLogger('pandasdmx').setLevel(log_level)

    def __getattr__(self, name):
        """Convenience methods."""
        try:
            # Provide resource_type as a positional argument, so that the
            # first positional argument to the convenience method is treated as
            # resource_id
            func = partial(self.get, Resource[name])
        except KeyError:
            raise AttributeError
        else:
            # Modify the docstring to explain the argument fixed by the
            # convenience method
            func.__doc__ = self.get.__doc__.replace(
                '.\n',
                f' with resource_type={repr(name)}.\n', 1)
            return func

    def __dir__(self):
        """Include convenience methods in dir()."""
        return super().__dir__() + [ep.name for ep in Resource]

    def clear_cache(self):
        self.cache.clear()

    @property
    def timeout(self):
        return self.session.timeout

    @timeout.setter
    def timeout(self, value):
        self.session.timeout = value

    def series_keys(self, flow_id, use_cache=True):
        """Return all :class:`pandasdmx.model.SeriesKey` for *flow_id*.

        Returns
        -------
        list
        """
        # download an empty dataset with all available series keys
        return self.data(flow_id, params={'detail': 'serieskeysonly'},
                         use_cache=use_cache) \
                   .data[0] \
                   .series \
                   .keys()

    def _make_key(self, resource_type, resource_id, key, dsd):
        """Validate *key* if possible.

        If key is a dict, validate items against the DSD and construct the key
        string which becomes part of the URL. Otherwise, do nothing as key must
        be a str confirming to the REST API spec.
        """
        if not (resource_type == Resource.data and isinstance(key, dict)):
            return key, dsd

        # Select validation method based on agency capabilities
        if dsd:
            # DSD was provided
            pass
        elif self.source.supports[Resource.datastructure]:
            # Retrieve the DataStructureDefinition
            dsd = self.dataflow(resource_id, params=dict(references='all'),
                                use_cache=True) \
                      .dataflow[resource_id].structure

            if dsd.is_external_reference:
                # DataStructureDefinition was not retrieved with the Dataflow
                # query; retrieve it explicitly
                dsd = self.get(resource=dsd, use_cache=True) \
                          .structure[dsd.id]
        else:
            # Construct a DSD from the keys
            dsd = DataStructureDefinition.from_keys(
                self.series_keys(resource_id))

        # Make a ContentConstraint from the key
        cc = dsd.make_constraint(key)

        return cc.to_query_string(dsd), dsd

    def _request_from_args(self, params={}, headers={}, resource=None,
        resource_type=None, resource_id=None, force=False, provider=None,
        version=None, key=None,  dsd=None, validate=True):
        """Validate arguments and prepare pieces for a request."""

        # Base URL
        url_parts = [self.source.url]

        # Resource arguments

        try:
            if resource_type:
                resource_type = Resource[resource_type]
        except KeyError:
            raise ValueError(f'resource_type ({resource_type!r}) must be in '
                             + Resource.describe())

        if resource:
            # Resource object is given
            assert isinstance(resource, MaintainableArtefact)

            # Class of the object
            if resource_type:
                assert resource_type == Resource.from_obj(resource)
            else:
                resource_type = Resource.from_obj(resource)
            if resource_id:
                assert resource_id == resource.id, (
                    f'mismatch between resource_id={resource_id!r} and '
                    f'resource={resource!r}')
            else:
                resource_id = resource.id

        if not (force or self.source.supports[resource_type]):
            raise NotImplementedError(f'{self.source.id} does not support the'
                                      f'{resource_type!r} API endpoint; '
                                      'override using force=True')

        url_parts.append(resource_type.name)

        # Data provider ID to use in the URL
        if resource_type == Resource.data:
            # Requests for data do not specific an agency in the URL
            if provider is not None:
                warn(f"'provider' argument is redundant for {resource_type!r}")
            provider_id = None
        else:
            provider_id = provider if provider else getattr(self.source, 'id',
                                                            None)

        url_parts.extend([provider_id, resource_id])

        if not version and resource_type != Resource.data:
            url_parts.append('latest')

        if validate:
            # Make the key, and retain the DSD (if any) for use in parsing
            key, dsd = self._make_key(resource_type, resource_id, key, dsd)

        url_parts.append(key)

        # Assemble final URL
        url = '/'.join(filter(None, url_parts))

        # Parameters: set 'references' to sensible defaults
        parameters = params.copy() # avoid side effects
        if 'references' not in parameters:
            if resource_type in [Resource.dataflow, Resource.datastructure] \
                    and resource_id:
                parameters['references'] = 'all'
            elif resource_type == Resource.categoryscheme:
                parameters['references'] = 'parentsandsiblings'

        # Headers: use headers from source config if not given by the caller
        if not headers and self.source and resource_type:
            headers = self.source.headers.get(resource_type.name, {})

        return dsd, requests.Request('get', url, params=parameters,
                                headers=headers)

    def _request_from_url(self, url, params={}, headers={}, dsd=None):
        return requests.Request('get', url, params=params,
                                headers=headers)

    def get(self, resource_type=None, resource_id=None, tofile=None,
            use_cache=False, dry_run=False, dsd=None, **kwargs):
        """Retrieve SDMX data or metadata.

        (Meta)data is retrieved from the :attr:`source` of the current Request.
        The item(s) to retrieve can be specified in one of two ways:

        1. `resource_type`, `resource_id`: These give the type (see
           :class:`Resource`) and, optionally, ID of the item(s). If the
           `resource_id` is not given, all items of the given type are
           retrieved.
        2. a `resource` object, i.e. a :class:`MaintainableArtefact
           <pandasdmx.model.MaintainableArtefact>`: `resource_type` and
           `resource_id` are determined by the object's class and :attr:`id
           <pandasdmx.model.IdentifiableArtefact.id>` attribute, respectively.

        Data is retrieved with `resource_type='data'`. In this case, the
        optional keyword argument `key` can be used to constrain the data that
        is retrieved. Examples of the formats for `key`:

        1. ``{'GEO': ['EL', 'ES', 'IE']}``: :class:`dict` with dimension
           name(s) mapped to an iterable of allowable values.
        2. ``{'GEO': 'EL+ES+IE'}``: :class:`dict` with dimension name(s)
           mapped to strings joining allowable values with `'+'`, the logical
           'or' operator for SDMX web services.
        3. ``'....EL+ES+IE'``: :class:`str` in which ordered dimension values
           (some empty, ``''``) are joined with ``'.'``. Using this form
           requires knowledge of the dimension order in the target data
           `resource_id`; in the example, dimension 'GEO' is the fifth of five
           dimensions: ``'.'.join(['', '', '', '', 'EL+ES+IE'])``.
           :meth:`CubeRegion.to_query_string
           <pandasdmx.model.CubeRegion.to_query_string>` can also be used to
           create properly formatted strings.

        For formats 1 and 2, but not 3, the `key` argument is validated against
        the relevant :class:`DataStructureDefinition
        <pandasdmx.model.DataStructureDefinition>`, either given with the `dsd`
        keyword argument, or retrieved from the web service before the main
        query.

        For the optional `param` keyword argument, some useful parameters are:

        - 'startperiod', 'endperiod': restrict the time range of data to
          retrieve.
        - 'references': control which item(s) related to a metadata resource
          are retrieved, e.g. `references='parentsandsiblings'`.

        Parameters
        ----------
        resource_type : str or :class:`Resource`, optional
            Type of resource to retrieve.
        resource_id : str, optional
            ID of the resource to retrieve.
        tofile : str or :class:`~os.PathLike` or `file-like object`, optional
            File path or file-like to write SDMX data as it is recieved.
        use_cache : bool, optional
            If :obj:`True`, return a previously retrieved :class:`~.Message`
            from :attr:`cache`, or update the cache with a newly-retrieved
            Message.
        dry_run : bool, optional
            If :obj:`True`, prepare and return a :class:`requests.Request`
            object, but do not execute the query. The prepared URL and headers
            can be examined by inspecting the returned object.
        dsd : :class:`~.DataStructureDefinition`
            Existing object used to validate the `key` argument. If not
            provided, an additional query executed to retrieve a DSD in order
            to validate the `key`.
        **kwargs
            Other, optional parameters (below).

        Other Parameters
        ----------------
        force : bool
            If :obj:`True`, execute the query even if the :attr:`source` does
            not support queries for the given `resource_type`. Default:
            :obj:`False`.
        headers : dict
            HTTP headers. Given headers will overwrite instance-wide headers
            passed to the constructor. Default: :obj:`None` to use the default
            headers of the :attr:`source`.
        key : str or dict
            For queries with `resource_type='data'`. :class:`str` values are
            not validated; :class:`dict` values are validated using
            :meth:`~.DataStructureDefinition.make_constraint`.
        params : dict
            Query parameters. The `SDMX REST web service guidelines <https://\
            github.com/sdmx-twg/sdmx-rest/tree/master/v2_1/ws/rest/docs>`_
            describe parameters and allowable values for different queries.
            `params` is not validated before the query is executed.
        provider : str
            ID of the agency providing the data or metadata. Default:
            ID of the :attr:`source` agency.

            An SDMX web service is a ‘data source’ operated by a specific,
            ‘source’ agency. A web service may host data or metadata originally
            published by one or more ‘provider’ agencies. Many sources are also
            providers. Other agencies—e.g. the SDMX Global Registry—simply
            aggregate (meta)data from other providers, but do not providing any
            (meta)data themselves.
        resource : :class:`~.MaintainableArtefact` subclass
            Object to retrieve. If given, `resource_type` and `resource_id` are
            ignored.
        version : str
            :attr:`~.VersionableArtefact.version>` of a resource to retrieve.
            Default: the keyword 'latest'.

        Returns
        -------
        :class:`~.Message` or :class:`~requests.Request`
            The requested SDMX message or, if `dry_run` is :obj:`True`, the
            prepared request object.

        Raises
        ------
        NotImplementedError
            If the :attr:`source` does not support the given `resource_type`
            and `force` is not :obj:`True`.

        """
        # add dsd to the kwarts as it is needed by some adapters for 
        # structure-specific requests.
        # TODO: do not use **kwargs in the above function signature. Rather make kwargs explicit.
        kwargs.update(dsd=dsd)

        # Allow sources to modify request args
        # TODO this should occur after most processing, defaults, checking etc.
        #      are performed, so that core code does most of the work.
        if self.source:
            self.source.modify_request_args(kwargs)

        # Handle arguments
        if 'url' in kwargs:
            req = self._request_from_url(**kwargs)
        else:
            kwargs.update(dict(
                resource_type=resource_type,
                resource_id=resource_id,
            ))
            dsd, req = self._request_from_args(**kwargs)

        req = self.session.prepare_request(req)

        # Now get the SDMX message via HTTP
        logger.info('Requesting resource from %s', req.url)
        logger.info('with headers %s' % req.headers)

        # Try to get resource from memory cache if specified
        if use_cache:
            try:
                return self.cache[req.url]
            except KeyError:
                logger.info('Not found in cache')

        if dry_run:
            return req

        try:
            response = self.session.send(req)
            response.raise_for_status()
        except requests.exceptions.ConnectionError as e:
            raise e from None
        except requests.exceptions.HTTPError as e:
            # Convert a 501 response to a Python NotImplementedError
            if e.response.status_code == 501:
                raise NotImplementedError(
                    '{!r} endpoint at {}'.format(resource_type, e.request.url))
            else:
                raise

        # Maybe copy the response to file as it's received
        response_content = remote.ResponseIO(response, tee=tofile)

        # Select reader class
        content_type = response.headers.get('content-type', None)
        try:
            Reader = get_reader_for_content_type(content_type)
        except ValueError:
            try:
                response, response_content = self.source.handle_response(
                    response, response_content)
                content_type = response.headers.get('content-type', None)
                Reader = get_reader_for_content_type(content_type)
            except ValueError:
                raise ValueError("can't determine a reader for response "
                                 "content type: %s" % content_type)

        # Instantiate reader
        reader = Reader()

        # Parse the message 
        # Only if   SDMXML content is given, use any provided or auto-queried DSD.
        if 'json' in content_type:
            msg = reader.read_message(response_content)
        else:
            msg = reader.read_message(response_content,
                                  dsd=dsd)

        # Store the HTTP response with the message
        msg.response = response

        # Call the finish_message() hook
        msg = self.source.finish_message(msg, self, **kwargs)

        # store in memory cache if needed
        if use_cache:
            self.cache[req.url] = msg

        return msg

    def preview_data(self, flow_id, key={}):
        """Return a preview of data.

        For the Dataflow *flow_id*, return all series keys matching *key*.
        preview_data() uses a feature supported by some data providers that
        returns :class:`SeriesKeys <pandasdmx.model.SeriesKey>` without the
        corresponding :class:`Observations <pandasdmx.model.Observation>`.

        To count the number of series::

            keys = sdmx.Request('PROVIDER').preview_data('flow')
            len(keys)

        To get a :mod:`pandas` object containing the key values::

            keys_df = sdmx.to_pandas(keys)

        Parameters
        ----------
        flow_id : str
            Dataflow to preview.
        key : dict, optional
            Mapping of *dimension* to *values*, where *values* may be a
            '+'-delimited list of values. If given, only SeriesKeys that match
            *key* are returned. If not given, preview_data is equivalent to
            ``list(req.series_keys(flow_id))``.

        Returns
        -------
        list of :class:`SeriesKey <pandasdmx.model.SeriesKey>`
        """
        # Retrieve the series keys
        all_keys = self.series_keys(flow_id)

        if len(key):
            # Construct a DSD from the keys
            dsd = DataStructureDefinition.from_keys(all_keys)

            # Make a ContentConstraint from *key*
            cc = dsd.make_constraint(key)

            # Filter the keys
            return [k for k in all_keys if k in cc]
        else:
            # No key is provided
            return list(all_keys)


def read_sdmx(filename_or_obj, format=None, **kwargs):
    """Load a SDMX-ML or SDMX-JSON message from a file or file-like object.

    Parameters
    ----------
    filename_or_obj : str or :class:`~os.PathLike` or file
    format : 'XML' or 'JSON', optional

    Other Parameters
    ----------------
    dsd : :class:`~.DataStructureDefinition`
        For “structure-specific” `format`=``XML`` messages only.
    """
    import pandasdmx.reader.sdmxml
    import pandasdmx.reader.sdmxjson

    readers = {
        'XML': pandasdmx.reader.sdmxml.Reader,
        'JSON': pandasdmx.reader.sdmxjson.Reader,
        }

    if isinstance(filename_or_obj, str):
        filename_or_obj = Path(filename_or_obj)

    try:
        # Use the file extension to guess the reader
        reader = readers[filename_or_obj.suffix.lstrip('.').upper()]

        # Open the file
        obj = open(filename_or_obj, 'br')
    except KeyError:
        if format:
            reader = readers[format]
            obj = open(filename_or_obj, 'br')
        else:
            msg = ("cannot identify SDMX message format from file name "
                   f"'{filename_or_obj.name}'; use  format='...'")
            raise RuntimeError(msg)
    except AttributeError:
        # File is already open

        # Read a line and then return the cursor to the initial position
        pos = filename_or_obj.tell()
        first_line = filename_or_obj.readline().strip()
        filename_or_obj.seek(pos)

        if first_line.startswith('{'):
            reader = readers['JSON']
        elif first_line.startswith('<'):
            reader = readers['XML']
        else:
            msg = f"cannot infer SDMX message format from '{first_line[:5]}..'"
            raise RuntimeError(msg)

        obj = filename_or_obj

    return reader().read_message(obj, **kwargs)


def read_url(url, **kwargs):
    """Request a URL directly."""
    return Request().get(url=url, **kwargs)



"""Tests for experimental code using pandas objects for internal storage.

See pandasdmx.experimental for more information.
"""
from pandasdmx.experimental import DataSet as PandasDataSet
from pandasdmx.model import (
    AttributeValue,
    DataAttribute,
    DataSet,
    Key,
    Observation,
    )
import pytest


# Run the tests on both the standard DataSet class, and the experimental,
# PandasDataSet version
@pytest.mark.parametrize('DataSetType', [
    DataSet,
    pytest.param(PandasDataSet, marks=pytest.mark.experimental),
])
def test_add_obs(DataSetType):
    # Create a Key and Attributes
    key = Key(CURRENCY='NZD', CURRENCY_DENOM='EUR',
              TIME_PERIOD='2018-01-01')
    obs_status = DataAttribute(id='OBS_STATUS')
    attr = {'OBS_STATUS': AttributeValue(value_for=obs_status, value='A')}

    obs = []
    for day, value in enumerate([5, 6, 7]):
        key = key.copy(TIME_PERIOD='2018-01-{:02d}'.format(day))
        obs.append(Observation(dimension=key, value=value,
                               attached_attribute=attr))

    ds = DataSetType()
    ds.add_obs(obs)

    # PandasDataSet does not store Observation objects internally, but should
    # emit them when the .obs property is accessed
    assert all(a == b for a, b in zip(ds.obs, obs))



from . import Source as BaseSource


class Source(BaseSource):
    _id = 'ILO'

    def modify_request_args(self, kwargs):
        """Handle two limitations of ILO's REST service.

        1. Service returns SDMX-ML 2.0 by default, whereas pandaSDMX only
           supports SDMX-ML 2.1. Set ``?format=generic_2_1`` query parameter.
        2. The service does not support values 'parents', 'parentsandsiblings'
           (the pandaSDMX default), and 'all' for the ``references`` query
           parameter. Override the default with ``?references=none``.

           .. note:: Valid values are: none, parents, parentsandsiblings,
              children, descendants, all, or a specific structure reference
              such as 'codelist'.
        """
        super().modify_request_args(kwargs)

        kwargs.setdefault('params', {})
        kwargs['params'].setdefault('format', 'generic_2_1')
        kwargs['params'].setdefault('references', 'none')



import pandasdmx as sdmx
from pandasdmx import model

from . import MessageTest
from .data import specimen

import pytest


class Test_ESTAT_dsd_apro_mk_cola(MessageTest):
    path = MessageTest.path / 'ESTAT'
    filename = 'apro_mk_cola-structure.xml'

    def test_codelists_keys(self, msg):
        assert len(msg.codelist) == 6
        assert isinstance(msg.codelist.CL_GEO, model.Codelist)

    def test_codelist_name(self, msg):
        assert msg.codelist.CL_GEO.UK.name.en == 'United Kingdom'
        assert msg.codelist.CL_FREQ.name.en == 'FREQ'

    def test_code_cls(self, msg):
        assert isinstance(msg.codelist.CL_FREQ.D, model.Code)

    def test_writer(self, msg):
        cls_as_dfs = sdmx.to_pandas(msg.codelist)

        # Number of codes expected in each Codelist
        count = {
            'CL_FREQ': 6,
            'CL_GEO': 41,
            'CL_OBS_FLAG': 10,
            'CL_OBS_STATUS': 3,
            'CL_PRODMILK': 12,
            'CL_UNIT': 1,
            }

        assert all(len(df) == count[id] for id, df in cls_as_dfs.items())

    def test_urn(self, msg):
        """https://github.com/dr-leo/pandaSDMX/issue/154."""
        expected = ('urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure='
                    'ESTAT:DSD_apro_mk_cola(1.0)')

        # URN is parsed
        assert msg.structure.DSD_apro_mk_cola.urn == expected


class TestDSDCommon(MessageTest):
    path = MessageTest.path / 'SGR'
    filename = 'common-structure.xml'

    def test_codelists_keys(self, msg):
        assert len(msg.codelist) == 5
        assert isinstance(msg.codelist.CL_FREQ, model.Codelist)

    def test_codelist_name(self, msg):
        assert msg.codelist.CL_FREQ.D.name.en == 'Daily'

    def test_code_cls(self, msg):
        assert isinstance(msg.codelist.CL_FREQ.D, model.Code)

    def test_annotations(self, msg):
        code = msg.codelist.CL_FREQ.A
        anno_list = list(code.annotations)
        assert len(anno_list) == 1
        a = anno_list[0]
        assert isinstance(a, model.Annotation)
        assert a.text.en.startswith('It is')
        assert a.type == 'NOTE'


class TestECB_EXR1(MessageTest):
    path = MessageTest.path / 'ECB_EXR' / '1'
    filename = 'structure.xml'

    def test_uri(self, msg):
        """https://github.com/dr-leo/pandaSDMX/issue/154."""
        expected = 'https://www.ecb.europa.eu/vocabulary/stats/exr/1'
        # URI is parsed
        assert msg.structure['ECB_EXR1'].uri == expected


def test_exr_constraints():
    with specimen('1/structure-full.xml') as f:
        m = sdmx.read_sdmx(f)
    ECB_EXR1 = m.structure['ECB_EXR1']

    # Test DimensionDescriptor
    dd = ECB_EXR1.dimensions

    # Correct order
    assert dd[0].id == 'FREQ'

    # Correct number of dimensions
    assert len(dd.components) == 6

    # Dimensions can be retrieved by name; membership can be tested
    assert 'W' in dd.get('FREQ')

    # Similar tests for AttributeDescriptor
    ad = ECB_EXR1.attributes
    assert len(ad.components) == 24
    assert ad[-1].id == 'UNIT_MULT'
    assert '5' in ad.get('UNIT_MULT')

    pytest.xfail('constrained codes not implemented')  # TODO
    assert len(m._constrained_codes), 14

    assert 'W' not in m._constrained_codes.FREQ

    key = {'FREQ': ['W']}

    assert m.in_codes(key)

    assert not m.in_constraints(key, raise_error=False)

    with pytest.raises(ValueError):
        m.in_constraints(key)

    assert m.in_constraints({'CURRENCY': ['CHF']})

    # test with invalid key
    with pytest.raises(TypeError):
        m._in_constraints({'FREQ': 'A'})

    # structure writer with constraints
    out = sdmx.to_pandas(m)
    cl = out.codelist
    assert cl.shape == (3555, 2)

    # unconstrained codelists
    out = sdmx.to_pandas(m, constraint=False)
    cl = out.codelist
    assert cl.shape, (4177, 2)



from pandasdmx.source import add_source, list_sources, sources


def test_list_sources():
    source_ids = list_sources()
    assert len(source_ids) == 14

    # Listed alphabetically
    assert source_ids[0] == 'ABS'
    assert source_ids[-1] == 'WB'


def test_source_support():
    # Implicitly supported endpoint
    assert sources['ILO'].supports['categoryscheme']

    # Specifically unsupported endpoint
    assert not sources['ESTAT'].supports['categoryscheme']

    # Explicitly supported structure-specific data
    assert sources['INEGI'].supports['structure-specific data']


def test_add_source():
    profile = """{
        "id": "FOO",
        "name": "Demo source",
        "url": "https://example.org/sdmx"
        }"""
    add_source(profile)

    # JSON sources do not support metadata endpoints, by default
    profile2 = """{
        "id": "BAR",
        "data_content_type": "JSON",
        "name": "Demo source",
        "url": "https://example.org/sdmx"
        }"""
    add_source(profile2)
    assert not sources['BAR'].supports['datastructure']



from io import StringIO
import logging

import pandas as pd
import pandasdmx as sdmx
import pytest
import requests
from .data import specimen


def test_read_sdmx(tmp_path):
    # Copy the file to a temporary file with an urecognizable suffix
    target = tmp_path / 'foo.badsuffix'
    with specimen('flat.json', opened=False) as original:
        target.open('w').write(original.read_text())

    # Can't infer message type for unknown file extension
    exc = ("cannot identify SDMX message format from file name 'foo.badsuffix'"
           "; use  format='...'")
    with pytest.raises(RuntimeError, match=exc):
        sdmx.read_sdmx(target)

    # Using the format= kwarg suppresses the error
    sdmx.read_sdmx(target, format='JSON')

    # Format can be inferred from an already-open file without extension
    with specimen('flat.json') as f:
        sdmx.read_sdmx(f)

    # Exception raised when the file contents don't allow to guess the format
    bad_file = StringIO('#! neither XML nor JSON')
    exc = "cannot infer SDMX message format from '#! ne..'"
    with pytest.raises(RuntimeError, match=exc):
        sdmx.read_sdmx(bad_file)


def test_request():
    # Constructor
    r = sdmx.Request(log_level=logging.ERROR)

    # Invalid source name raise an exception
    with pytest.raises(ValueError):
        sdmx.Request('noagency')

    # Regular methods
    r.clear_cache()

    r.timeout = 300
    assert r.timeout == 300

    # dir() includes convenience methods for resource endpoints
    expected = {'cache', 'clear_cache', 'get', 'preview_data', 'series_keys',
                'session', 'source', 'timeout'}
    expected |= set(ep.name for ep in sdmx.Resource)
    assert set(filter(lambda s: not s.startswith('_'), dir(r))) == expected


def test_request_get_exceptions():
    """Tests of Request.get() that don't require remote data."""
    req = sdmx.Request('ESTAT')

    # Exception is raised on unrecognized arguments
    with pytest.raises(TypeError):
        req.get('datastructure', foo='bar')

    with pytest.raises(TypeError):
        sdmx.read_url('https://example.com', foo='bar')


@pytest.mark.remote_data
def test_request_get_args():
    req = sdmx.Request('ESTAT')

    # Request._make_key accepts '+'-separated values
    args = dict(resource_id='une_rt_a', key={'GEO': 'EL+ES+IE'},
                params={'startPeriod': '2007'}, dry_run=True, use_cache=True)
    # Store the URL
    url = req.data(**args).url

    # Using an iterable of key values gives the same URL
    args['key'] = {'GEO': ['EL', 'ES', 'IE']}
    assert req.data(**args).url == url

    # Using a direct string for a key gives the same URL
    args['key'] = '....EL+ES+IE'  # No specified values for first 4 dimensions
    assert req.data(**args).url == url

    # Giving 'provider' is redundant for a data request, causes a warning
    with pytest.warns(UserWarning, match="'provider' argument is redundant"):
        req.data('une_rt_a', key={'GEO': 'EL+ES+IE'},
                 params={'startPeriod': '2007'},
                 provider='ESTAT')

    # Using an unknown endpoint is an exception
    with pytest.raises(ValueError):
        req.get('badendpoint', 'id')

    # TODO test req.get(obj) with IdentifiableArtefact subclasses


@pytest.mark.remote_data
def test_read_url():
    # URL can be queried without instantiating Request
    sdmx.read_url('http://sdw-wsrest.ecb.int/service/datastructure/ECB/'
                  'ECB_EXR1/latest?references=all')


@pytest.mark.remote_data
def test_request_preview_data():
    req = sdmx.Request('ECB')

    # List of keys can be retrieved
    keys = req.preview_data('EXR')
    assert isinstance(keys, list)

    # Count of keys can be determined
    assert len(keys) > 1000

    # A filter can be provided, resulting in fewer keys
    keys = req.preview_data('EXR', {'CURRENCY': 'CAD+CHF+CNY'})
    assert len(keys) == 24

    # Result can be converted to pandas object
    keys_pd = sdmx.to_pandas(keys)
    assert isinstance(keys_pd, pd.DataFrame)
    assert len(keys_pd) == 24


def test_custom_session():
    # provided session is stored in Request instance
    sess = requests.Session()
    req = sdmx.Request('ECB', session=sess)
    assert req.session is sess
    
    # session_opts are not allowed when session is given
    with pytest.raises(TypeError):
        req = sdmx.Request('ECB', session=sess, foo='bar')
    
    


"""Specimens and other data for pandaSDMX tests.

This directory contains files with SDMX data and structure messages, in both
SDMX-ML ('.xml') and JSON ('.json') formats.

These files can be:
- retrieved with :meth:`specimen`,
- used to parametrize test methods with :meth:`test_data`, and
- compared to expected :meth:`to_pandas` output using :meth:`expected_data`.

The files are:

- Arranged in directories with names matching particular sources in
  :file:`sources.json`.

- Named with:

  - Certain keywords:

    - '-structure': a structure message, often associated with a file with a
      similar name containing a data message.
    - 'ts': time-series data, i.e. with a TimeDimensions at the level of
      individual Observations.
    - 'xs': cross-sectional data arranged in other ways.
    - 'flat': flat DataSets with all Dimensions at the Observation level.
    - 'ss': structure-specific data messages.

  - In some cases, the query string or data flow/structure ID as the file name.

  - Hyphens '-' instead of underscores '_'.

"""
from contextlib import contextmanager
from pathlib import Path

import pandas as pd


BASE_PATH = Path(__file__).parent

# List of specimen files.
# Each is a tuple: (path, format (xml|json), kind (data|structure))
TEST_FILES = [
    (BASE_PATH / 'INSEE' / 'IPI-2010-A21.xml', 'xml', 'data'),
]

# XML data files for the ECB exchange rate data flow
for path in (BASE_PATH / 'ECB_EXR').rglob('*.xml'):
    kind = 'data'
    if 'structure' in path.name or 'common' in path.name:
        kind = 'structure'
    TEST_FILES.append((path, 'xml', kind))

# JSON data files for the ECB exchange rate data flow
for fp in (BASE_PATH / 'ECB_EXR').rglob('*.json'):
    TEST_FILES.append((fp, 'json', 'data'))

# Miscellaneous XML data files
TEST_FILES.append(
    (BASE_PATH / 'ESTAT' / 'footer.xml', 'xml', 'data'),
    )


# Miscellaneous XML structure files
TEST_FILES.extend(
    (BASE_PATH.joinpath(*parts), 'xml', 'structure') for parts in [
        ('ECB', 'orgscheme.xml'),
        ('ESTAT', 'apro_mk_cola-structure.xml'),

        # Manually reduced subset of the response for this DSD. Test for
        # <str:CubeRegion> containing both <com:KeyValue> and <com:Attribute>
        ('IMF', 'ECOFIN_DSD-structure.xml'),

        ('INSEE', 'CNA-2010-CONSO-SI-A17-structure.xml'),
        ('INSEE', 'dataflow.xml'),
        ('INSEE', 'IPI-2010-A21-structure.xml'),
        ('UNSD', 'codelist_partial.xml'),
        ('SGR', 'common-structure.xml'),
        ])

# Expected to_pandas() results for data files above; see expected_data()
# - Keys are the file name (above) with '.' -> '-': 'foo.xml' -> 'foo-xml'
# - Data is stored in expected/{KEY}.txt
# - Values are either argument to pd.read_csv(); or a dict(use='other-key'),
#   in which case the info for other-key is used instead.
EXPECTED = {
    'ng-flat-xml': dict(index_col=[0, 1, 2, 3, 4, 5]),
    'ng-ts-gf-xml': dict(use='ng-flat-xml'),
    'ng-ts-xml': dict(use='ng-flat-xml'),
    'ng-xs-xml': dict(index_col=[0, 1, 2, 3, 4, 5]),
    # Excluded: this file contains two DataSets, and expected_data() currently
    # only supports specimens with one DataSet
    # 'action-delete-json': dict(header=[0, 1, 2, 3, 4]),
    'xs-json': dict(index_col=[0, 1, 2, 3, 4, 5]),
    'flat-json': dict(index_col=[0, 1, 2, 3, 4, 5]),
    'ts-json': dict(use='flat-json'),
}


def test_files(format=None, kind=None):
    """Generate a sequence of test file paths matching criteria.

    The return value should be passed as kwargs to pytest.mark.parametrize():

        pytest.mark.parametrize('argname', **test_files(…))

    """
    result = dict(argvalues=[], ids=[])
    for path, f, k in TEST_FILES:
        if (format and format != f) or (kind and kind != k):
            continue
        result['argvalues'].append(path)
        result['ids'].append(path.name)
    return result


@contextmanager
def specimen(pattern='', opened=True):
    """Open the test specimen file with *pattern* in the name."""
    for path, f, k in TEST_FILES:
        if path.match('*' + pattern + '*'):
            yield open(path) if opened else path
            return
    raise ValueError(pattern)


def expected_data(path):
    """Return the expected to_pandas() result for *path*."""
    try:
        key = path.name.replace('.', '-')
        info = EXPECTED[key]
        if 'use' in info:
            # Use the same expected data as another file
            key = info['use']
            info = EXPECTED[key]
    except KeyError:
        return None

    args = dict(sep=r'\s+', index_col=[0], header=[0])
    args.update(info)

    expected_path = (BASE_PATH / 'expected' / key).with_suffix('.txt')
    result = pd.read_csv(expected_path, **args)

    # A series; unwrap
    if set(result.columns) == {'value'}:
        result = result['value']

    return result



from enum import Enum
from importlib import import_module, resources
import json
from typing import (
    Any,
    Dict,
    Union,
    )

from pandasdmx.model import DataStructureDefinition
from pandasdmx.util import BaseModel, Resource
from pydantic import validator


sources = {}

DataContentType = Enum('DataContentType', 'XML JSON')


class Source(BaseModel):
    """SDMX-IM RESTDatasource.

    This class describes the location and features supported by an SDMX data
    source. Subclasses may override the hooks in order to handle specific
    features of different REST web services:

    .. autosummary::
       handle_response
       finish_message
       modify_request_args

    """
    #: ID of the data source
    id: str

    #: Base URL for queries
    url: str

    #: Human-readable name of the data source
    name: str

    headers: Dict[str, Any] = {}

    #: :class:`pandasdmx.util.DataContentType` indicating the type of data
    #: returned by the source.
    data_content_type: DataContentType = DataContentType.XML

    #: Mapping from :class:`Resource <pandasdmx.util.Resource>` to
    #: :class:`bool` indicating support for SDMX REST API features. Two
    #: additional keys are valid:
    #:
    #: - ``'preview'=True`` if the source supports ``?detail=serieskeysonly``.
    #:   See :meth:`preview_data <pandasdmx.Request.preview_data>`.
    #: - ``'structure-specific data'=True`` if the source can return structure-
    #:   specific data messages.
    supports: Dict[Union[str, Resource], bool] = {Resource.data: True}

    @classmethod
    def from_dict(cls, info):
        return cls(**info)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set default supported features
        for feature in list(Resource) + ['preview', 'structure-specific data']:
            self.supports.setdefault(
                feature, self.data_content_type == DataContentType.XML)

    # Hooks
    def handle_response(self, response, content):
        """Handle response content of unknown type.

        This hook is called by :meth:`pandasdmx.Request.get` *only* when
        the `content` cannot be parsed as XML or JSON.

        See :meth:`ESTAT <pandasdmx.source.estat.Source.handle_response>` and
        :meth:`SGR <pandasdmx.source.sgr.Source.handle_response>` for example
        implementations.
        """
        return response, content

    def finish_message(self, message, request, **kwargs):
        """Postprocess retrieved message.

        This hook is called by :meth:`pandasdmx.Request.get` after a
        :class:`pandasdmx.message.Message` object has been successfully
        parsed from the query response.

        See :meth:`ESTAT <pandasdmx.source.estat.Source.finish_message>` for an
        example implementation.
        """
        return message

    def modify_request_args(self, kwargs):
        """Modify arguments used to build query URL.

        This hook is called by :meth:`pandasdmx.Request.get` to modify the
        keyword arguments before the query URL is built.

        The default implementation handles requests for 'structure-specific
        data' by adding an HTTP 'Accepts:' header when a 'dsd' is supplied as
        one of the `kwargs`.

        See :meth:`SGR <pandasdmx.source.sgr.Source.modify_request_args>` for
        an example override.

        Returns
        -------
        None
        """
        if self.data_content_type is DataContentType.XML:
            dsd = kwargs.get('dsd', None)
            if isinstance(dsd, DataStructureDefinition):
                kwargs.setdefault('headers', {})
                kwargs['headers'].setdefault(
                    'Accept',
                    'application/vnd.sdmx.structurespecificdata+xml;'
                    'version=2.1')

    @validator('id')
    def _validate_id(cls, value):
        assert getattr(cls, '_id', value) == value
        return value

    @validator('data_content_type', pre=True)
    def _validate_dct(cls, value):
        if isinstance(value, DataContentType):
            return value
        else:
            return DataContentType[value]


class _NoSource(Source):
    id = ''
    url = ''
    name = ''


NoSource = _NoSource()


def add_source(info, id=None, override=False, **kwargs):
    """Add a new data source.

    The *info* expected is in JSON format:

    .. code-block:: json

        {
          "id": "ESTAT",
          "documentation": "http://data.un.org/Host.aspx?Content=API",
          "url": "http://ec.europa.eu/eurostat/SDMX/diss-web/rest",
          "name": "Eurostat",
          "supported": {"codelist": false, "preview": true}
        }

    …with unspecified values using the defaults; see
    :class:`Source`.

    Parameters
    ----------
    info : dict-like
        String containing JSON information about a data source.
    id : str
        Identifier for the new datasource. If :obj:`None` (default), then
        `info['id']` is used.
    override : bool
        If :obj:`True`, replace any existing data source with *id*.
        Otherwise, raise :class:`ValueError`.
    **kwargs
        Optional callbacks for *handle_response* and *finish_message* hooks.

    """
    if isinstance(info, str):
        info = json.loads(info)

    id = info['id'] if id is None else id

    info.update(kwargs)

    if id in sources:
        raise ValueError("Data source '%s' already defined; use override=True",
                         id)

    # Maybe import a subclass that defines a hook
    SourceClass = Source
    try:
        mod = import_module('.' + id.lower(), 'pandasdmx.source')
    except ImportError:
        pass
    else:
        SourceClass = mod.Source

    sources[id] = SourceClass.from_dict(info)


def list_sources():
    """Return a sorted list of valid source IDs.

    These can be used to create :class:`Request` instances.
    """
    return sorted(sources.keys())


def load_package_sources():
    """Discover all sources listed in ``sources.json``."""
    with resources.open_binary('pandasdmx', 'sources.json') as f:
        for info in json.load(f):
            add_source(info)


load_package_sources()



"""Speed and memory usage tests."""
from pandasdmx.model import (
    AttributeValue,
    DataAttribute,
    DataStructureDefinition,
    )


def test_refcount():
    # Component (subclasses) created outside of a DataStructureDefintion
    da1 = DataAttribute(id='foo')
    da2 = DataAttribute(id='foo')

    assert id(da1) != id(da2)
    assert da1 is not da2

    # Retrieving attributes from a DataStructureDefintion results in references
    # to the same object
    dsd = DataStructureDefinition()

    da3 = dsd.attribute('foo')
    da4 = dsd.attribute('foo')

    assert id(da3) == id(da4)
    assert da3 is da4
    assert len(dsd.attributes) == 1

    # Creating an AttributeValue referencing a DataAttribute outside a DSD
    av1 = AttributeValue(value='baz', value_for=DataAttribute(id='foo'))
    assert not any([av1.value_for is da for da in (da1, da2, da3)])

    # Same, using a DSD
    av2 = AttributeValue(value='baz', value_for='foo', dsd=dsd)
    assert av2.value_for is da3



import re

from . import Source as BaseSource


re_500 = re.compile(r"(An error has occurred)\.")


class Source(BaseSource):
    _id = 'ABS'

    def handle_response(self, response, content):
        """Handle ABS' own text/html error page for some endpoints."""
        ctype = response.headers.get('content-type', '')
        if 'text/html' in ctype:
            buf = ''
            while True:
                chunk = content.read().decode()
                if len(chunk) == 0:
                    break

                buf += chunk

                match = re_500.search(buf)
                if match:
                    # Overwrite the original response
                    response.reason, = match.groups()
                    response.status_code = 500
                    response.raise_for_status()

        raise NotImplementedError



from pandasdmx.util import BaseModel, DictLike, validate_dictlike
import pydantic
from pydantic import StrictStr
import pytest


def test_dictlike():
    dl = DictLike()

    # Set by item name
    dl['TIME_PERIOD'] = 3
    dl['CURRENCY'] = 'USD'

    # Access by attribute name
    assert dl.TIME_PERIOD == 3

    # Access by item index
    assert dl[1] == 'USD'

    # Access beyond index
    with pytest.raises(KeyError):
        dl['FOO']

    with pytest.raises(IndexError):
        dl[2]

    with pytest.raises(AttributeError):
        dl.FOO


def test_dictlike_anno():
    @validate_dictlike('items')
    class Foo(BaseModel):
        items: DictLike[StrictStr, int] = DictLike()

    f = Foo()
    assert type(f.items) == DictLike

    # Can be set with DictLike
    f.items = DictLike(a=1, b=2)
    assert type(f.items) == DictLike

    # Can be set with dict()
    f.items = {'a': 1, 'b': 2}
    assert type(f.items) == DictLike

    # Type checking on creation
    with pytest.raises(pydantic.ValidationError):
        f = Foo(items={1: 'a'})

    # Type checking on assignment
    f = Foo()
    with pytest.raises(pydantic.ValidationError):
        f.items = {1: 'a'}

    # Type checking on setting elements
    f = Foo(items={'a': 1})
    with pytest.raises(pydantic.ValidationError):
        f.items[123] = 456

    # commented: this does not work, since validate_dictlike does not operate
    # until initial values are assigned to the field
    # f = Foo()
    # with pytest.raises(pydantic.ValidationError):
    #     f.items[123] = 456

    # Use validate_dictlike() twice
    @validate_dictlike('elems')
    class Bar(BaseModel):
        elems: DictLike[StrictStr, float] = DictLike()



"""Tests for pandasdmx/writer.py."""
import pandas as pd
import pytest
from pytest import raises

import pandasdmx as sdmx
from pandasdmx.model import TimeDimension

from . import assert_pd_equal
from .data import expected_data, specimen, test_files


# file name → (exception raised, exception message, comment/reason)
ssds = ('Reading StructureSpecificDataSet does not distinguish between attrs '
        'and dimension values.')

file_marks = {
    'exr-action-delete.json': (
        AssertionError,
        "Expected type <class 'pandas.core.frame.DataFrame'>, found <class "
        " 'list'> instead",
        'Message contains two DataSets; test infrastructure currently handles '
        'only messages with a single DataSet.'),
    'ECB/EXR/ng-ts-ss.xml':
        (AssertionError, 'Series.index are different', ssds),
    'ECB/EXR/ng-flat-ss.xml':
        (AssertionError, 'Series.index are different', ssds),
    'ECB/EXR/ng-xs-ss.xml':
        (AssertionError, 'Series.index are different', ssds),
    'ECB/EXR/ng-ts-gf-ss.xml':
        (AssertionError, 'Series.index are different', ssds),
    }


def pytest_generate_tests(metafunc):
    if 'data_path' in metafunc.fixturenames:
        params = []
        tf = test_files(kind='data')
        for value, id in zip(tf['argvalues'], tf['ids']):
            kwargs = dict(id=id)
            for cond, info in file_marks.items():
                if cond in str(value):
                    kwargs['marks'] = pytest.mark.skip(reason=info[2])
                    break

            params.append(pytest.param(value, **kwargs))

        metafunc.parametrize('data_path', params)


def test_write_data_arguments():
    msg = sdmx.read_sdmx(test_files(kind='data')['argvalues'][0])

    # Attributes must be a string
    with raises(TypeError):
        sdmx.to_pandas(msg, attributes=2)
        msg.to_pandas(attributes=2)
        

    # Attributes must contain only 'dgso'
    with raises(ValueError):
        sdmx.to_pandas(msg, attributes='foobarbaz')
        msg.to_pandas(attributes='foobarbaz')


def test_write_data(data_path):
    msg = sdmx.read_sdmx(data_path)

    result = sdmx.to_pandas(msg)

    expected = expected_data(data_path)
    if expected is not None:
        print(expected, result, sep='\n')
    assert_pd_equal(expected, result)

    # TODO incomplete
    assert isinstance(result, (pd.Series, pd.DataFrame, list)), type(result)


@pytest.mark.parametrize('path', **test_files(kind='data'))
def test_write_data_attributes(path):
    msg = sdmx.read_sdmx(path)

    result = sdmx.to_pandas(msg, attributes='osgd')
    # TODO incomplete
    assert isinstance(result, (pd.Series, pd.DataFrame, list)), type(result)


def test_write_agencyscheme():
    # Convert an agency scheme
    with specimen('ECB/orgscheme.xml') as f:
        msg = sdmx.read_sdmx(f)
        data = sdmx.to_pandas(msg)

    assert data['organisation_scheme']['AGENCIES']['ESTAT'] == 'Eurostat'

    # to_pandas only returns keys for non-empty attributes of StructureMessage
    # https://github.com/dr-leo/pandaSDMX/issues/90
    assert set(data.keys()) == {'organisation_scheme'}

    # Attribute access works
    assert data.organisation_scheme.AGENCIES.ESTAT == 'Eurostat'

    with pytest.raises(AttributeError):
        data.codelist
    with pytest.raises(AttributeError):
        data.dataflow
    with pytest.raises(AttributeError):
        data.structure

@pytest.mark.filterwarnings('ignore::DeprecationWarning')
def test_write_categoryscheme():
    with specimen('IPI-2010-A21-structure.xml') as f:
        msg = sdmx.read_sdmx(f)
        print(msg.category_scheme)
        data = sdmx.to_pandas(msg)
        same_data = msg.write()
        assert type(data) is type(same_data)

    cs = data['category_scheme']['CLASSEMENT_DATAFLOWS']

    assert (cs.loc['COMPTA-NAT', 'name']
            == 'National accounts (GDP, consumption...)')

    # Children appear
    assert cs.loc['CNA-PIB-2005', 'parent'] == 'CNA-PIB'


def test_write_codelist():
    # Retrieve codelists from a test specimen and convert to pandas
    with specimen('common-structure.xml') as f:
        dsd_common = sdmx.read_sdmx(f)
    codelists = sdmx.to_pandas(dsd_common)['codelist']

    # File contains 5 code lists
    assert len(codelists) == 5

    # Code lists have expected number of items
    assert len(codelists['CL_FREQ']) == 8

    # Items names can be retrieved by ID
    freq = codelists['CL_FREQ']
    assert freq['A'] == 'Annual'

    # Non-hierarchical code list has a string name
    assert freq.name == 'Code list for Frequency (FREQ)'

    # Hierarchical code list
    with specimen('codelist_partial.xml') as f:
        msg = sdmx.read_sdmx(f)

    # Convert single codelist
    CL_AREA = sdmx.to_pandas(msg.codelist['CL_AREA'])

    # Hierichical list has a 'parent' column; parent of Africa is the World
    assert CL_AREA.loc['002', 'parent'] == '001'

    # Pandas features can be used to merge parent names
    area_hierarchy = pd.merge(CL_AREA, CL_AREA,
                              how='left', left_on='parent', right_index=True,
                              suffixes=('', '_parent'))
    assert area_hierarchy.loc['002', 'name_parent'] == 'World'


def test_write_conceptscheme():
    with specimen('common-structure.xml') as f:
        msg = sdmx.read_sdmx(f)
        data = sdmx.to_pandas(msg)

    cdc = data['concept_scheme']['CROSS_DOMAIN_CONCEPTS']
    assert cdc.loc['UNIT_MEASURE', 'name'] == 'Unit of Measure'


def test_write_dataflow():
    # Read the INSEE dataflow definition
    with specimen('INSEE/dataflow') as f:
        msg = sdmx.read_sdmx(f)

    # Convert to pandas
    result = sdmx.to_pandas(msg, include='dataflow')

    # Number of Dataflows described in the file
    assert len(result['dataflow']) == 663

    # ID and names of first Dataflows
    mbop = 'Monthly Balance of Payments - '
    expected = pd.Series({
        'ACT-TRIM-ANC': 'Activity by sex and age - Quarterly series',
        'BPM6-CCAPITAL': '{}Capital account'.format(mbop),
        'BPM6-CFINANCIER': '{}Financial account'.format(mbop),
        'BPM6-CTRANSACTION': '{}Current transactions account'.format(mbop),
        'BPM6-TOTAL': '{}Overall total and main headings'.format(mbop),
        })
    assert_pd_equal(result['dataflow'].head(), expected)


def test_write_dataset_datetime():
    """Test datetime arguments to write_dataset()."""
    # Load structure
    with specimen('IPI-2010-A21-structure.xml') as f:
        dsd = sdmx.read_sdmx(f).structure['IPI-2010-A21']
        TIME_PERIOD = dsd.dimensions.get('TIME_PERIOD')
        FREQ = dsd.dimensions.get('FREQ')

    assert isinstance(TIME_PERIOD, TimeDimension)

    # Load data, two ways
    with specimen('IPI-2010-A21.xml') as f:
        msg = sdmx.read_sdmx(f, dsd=dsd)
        ds = msg.data[0]
    with specimen('IPI-2010-A21.xml') as f:
        msg_no_structure = sdmx.read_sdmx(f)

    other_dims = list(filter(lambda n: n != 'TIME_PERIOD',
                             [d.id for d in dsd.dimensions.components]))

    def expected(df, axis=0, cls=pd.DatetimeIndex):
        axes = ['index', 'columns'] if axis else ['columns', 'index']
        assert getattr(df, axes[0]).names == other_dims
        assert isinstance(getattr(df, axes[1]), cls)

    # Write with datetime=str
    df = sdmx.to_pandas(ds, datetime='TIME_PERIOD')
    expected(df)

    # Write with datetime=Dimension instance
    df = sdmx.to_pandas(ds, datetime=TIME_PERIOD)
    expected(df)

    # Write with datetime=True fails because the data message contains no
    # actual structure information
    with pytest.raises(ValueError, match=r'no TimeDimension in \[.*\]'):
        sdmx.to_pandas(msg_no_structure, datetime=True)
    with pytest.raises(ValueError, match=r'no TimeDimension in \[.*\]'):
        sdmx.to_pandas(msg_no_structure.data[0], datetime=True)

    # DataMessage parsed with a DSD allows write_dataset to infer the
    # TimeDimension
    df = sdmx.to_pandas(msg, datetime=True)
    expected(df)
    # Same for DataSet
    df = sdmx.to_pandas(ds, datetime=True)
    expected(df)

    # As above, with axis=1
    df = sdmx.to_pandas(ds, datetime=dict(dim='TIME_PERIOD', axis=1))
    expected(df, axis=1)
    df = sdmx.to_pandas(ds, datetime=dict(dim=TIME_PERIOD, axis=1))
    expected(df, axis=1)
    ds.structured_by = dsd
    df = sdmx.to_pandas(ds, datetime=dict(axis=1))
    expected(df, axis=1)
    df = sdmx.to_pandas(msg, datetime=dict(axis=1))
    expected(df, axis=1)

    # Write with freq='M' works
    df = sdmx.to_pandas(ds, datetime=dict(dim='TIME_PERIOD', freq='M'))
    expected(df, cls=pd.PeriodIndex)

    # Write with freq='A' works
    df = sdmx.to_pandas(ds, datetime=dict(dim='TIME_PERIOD', freq='A'))
    expected(df, cls=pd.PeriodIndex)
    # …but the index is not unique, because month information was discarded
    assert not df.index.is_unique

    # Write specifying the FREQ dimension by name fails
    with pytest.raises(ValueError, match='cannot convert to PeriodIndex with '
                       r"non-unique freq=\['A', 'M'\]"):
        sdmx.to_pandas(ds, datetime=dict(dim='TIME_PERIOD', freq='FREQ'))

    # Remove non-monthly obs
    # TODO use a constraint, when this is supported
    ds.obs = list(filter(lambda o: o.key.FREQ != 'A', ds.obs))

    # Now specifying the dimension by name works
    df = sdmx.to_pandas(ds, datetime=dict(dim='TIME_PERIOD', freq='FREQ'))

    # and FREQ is no longer in the columns index
    other_dims.pop(other_dims.index('FREQ'))
    expected(df, cls=pd.PeriodIndex)

    # Specifying a Dimension works
    df = sdmx.to_pandas(ds, datetime=dict(dim=TIME_PERIOD, freq=FREQ))
    expected(df, cls=pd.PeriodIndex)

    # As above, using DSD attached to the DataMessage
    df = sdmx.to_pandas(msg, datetime=dict(dim=TIME_PERIOD, freq='FREQ'))
    expected(df, cls=pd.PeriodIndex)

    # Invalid arguments
    with pytest.raises(ValueError, match='X'):
        sdmx.to_pandas(msg, datetime=dict(dim=TIME_PERIOD, freq='X'))
    with pytest.raises(ValueError, match='foo'):
        sdmx.to_pandas(ds, datetime=dict(foo='bar'))
    with pytest.raises(ValueError, match='43'):
        sdmx.to_pandas(ds, datetime=43)


@pytest.mark.parametrize('path', **test_files(kind='structure'))
def test_writer_structure(path):
    msg = sdmx.read_sdmx(path)

    sdmx.to_pandas(msg)

    # TODO test contents


@pytest.mark.remote_data
def test_write_constraint():
    """'constraint' argument to writer.write_dataset."""
    with specimen('ng-ts.xml') as f:
        msg = sdmx.read_sdmx(f)

    # Fetch the message's DSD
    assert msg.structure.is_external_reference
    # NB the speciment included in tests/data has 'ECB_EXR_NG' as the
    #    data structure ID; but a query against the web service gives
    #    'ECB_EXR1' for the same data structure.
    id = 'ECB_EXR1'
    dsd = sdmx.Request(msg.structure.maintainer.id) \
              .get('datastructure', id) \
              .structure[id]

    # Create a ContentConstraint
    cc = dsd.make_constraint({'CURRENCY': 'JPY+USD'})

    # Write the message without constraint
    s1 = sdmx.to_pandas(msg)
    assert len(s1) == 12
    assert set(s1.index.to_frame()['CURRENCY']) == {'CHF', 'GBP', 'JPY', 'USD'}

    # Writing using constraint produces a fewer items; only those matching the
    # constraint
    s2 = sdmx.to_pandas(msg, constraint=cc)
    assert len(s2) == 6
    assert set(s2.index.to_frame()['CURRENCY']) == {'JPY', 'USD'}



import pandasdmx as sdmx
from pandasdmx.message import (
    Header,
    DataMessage,
    )
from pandasdmx.model import (
    AttributeValue,
    DataAttribute,
    DataSet,
    Key,
    Observation,
    )

from . import assert_pd_equal
from .data import specimen


def test_flat():
    # Create a bare Message
    msg = DataMessage()

    # Recreate the content from exr-flat.json
    header = Header(
        id='62b5f19d-f1c9-495d-8446-a3661ed24753',
        prepared='2012-11-29T08:40:26Z',
        sender='ECB',
        )
    msg.header = header

    ds = DataSet()

    # Create a Key and attributes
    key = Key(FREQ='D', CURRENCY='NZD', CURRENCY_DENOM='EUR', EXR_TYPE='SP00',
              EXR_SUFFIX='A', TIME_PERIOD='2013-01-18')
    obs_status = DataAttribute(id='OBS_STATUS')
    attr = {'OBS_STATUS': AttributeValue(value_for=obs_status, value='A')}

    ds.obs.append(Observation(dimension=key, value=1.5931,
                  attached_attribute=attr))

    key = key.copy(TIME_PERIOD='2013-01-21')
    ds.obs.append(Observation(dimension=key, value=1.5925,
                  attached_attribute=attr))

    key = key.copy(CURRENCY='RUB', TIME_PERIOD='2013-01-18')
    ds.obs.append(Observation(dimension=key, value=40.3426,
                  attached_attribute=attr))

    key = key.copy(TIME_PERIOD='2013-01-21')
    ds.obs.append(Observation(dimension=key, value=40.3000,
                  attached_attribute=attr))

    msg.data.append(ds)

    # Write to pd.Dataframe
    df1 = sdmx.to_pandas(msg)

    with specimen('flat.json') as f:
        ref = sdmx.read_sdmx(f)
    df2 = sdmx.to_pandas(ref)

    assert_pd_equal(df1, df2)


def test_bare_series():
    with specimen('ng-ts.xml') as f:
        sdmx.read_sdmx(f)

    # TODO generate the series and observations



from io import BufferedIOBase, BytesIO
import logging
from warnings import warn

import requests

try:
    from requests_cache import CachedSession as MaybeCachedSession
except ImportError:  # pragma: no cover
    warn('optional dependency requests_cache is not installed; cache options '
         'to Session() have no effect', RuntimeWarning)
    from requests import Session as MaybeCachedSession


logger = logging.getLogger(__name__)


class Session(MaybeCachedSession):
    """:class:`requests.Session` subclass with optional caching.

    If requests_cache is installed, this class caches responses.
    """

    def __init__(self, timeout=30.1, proxies=None, stream=False, **kwargs):

        if MaybeCachedSession is not requests.Session:
            # Using requests_cache.CachedSession

            # No cache keyword arguments supplied = don't use the cache
            disabled = set(kwargs.keys()) <= {'get_footer_url'}

            if disabled:
                # Avoid creating any file
                kwargs['backend'] = 'memory'

            super(Session, self).__init__(**kwargs)

            # Overwrite value from requests_cache.CachedSession.__init__()
            self._is_cache_disabled = disabled
        elif len(kwargs):
            raise ValueError('Cache arguments have no effect without '
                             'requests_session: %s' % kwargs)
        else:
            # Plain requests.Session
            super(Session, self).__init__()

        # Overwrite values from requests.Session.__init__()
        self.proxies = proxies
        self.timeout = timeout
        self.stream = stream


class ResponseIO(BufferedIOBase):
    """Read  data from a :class:`requests.Response` object, into an in-memory or on-disk file 
    and expose it as a file-like object.

    :class:`ResponseIO` wraps a :class:`requests.Response` object's 'content'
    attribute, providing a file-like object from which bytes can be
    :meth:`read` incrementally.

    Parameters
    ----------
    response : :class:`requests.Response`
        HTTP response to wrap.
    tee : binary, writable :py:class:`io.BufferedIOBasè`, defaults to io.BytesIO()
        *tee* is exposed as *self.tee* and not closed explicitly.
    """

    def __init__(self, response, tee=None):
        self.response = response
        if tee is None:
            tee = BytesIO()
        # If tee is a file-like object or tempfile, then use it as cache
        if isinstance(tee, BufferedIOBase) or hasattr(tee, 'file'):
            self.tee = tee
        else:
            # So tee must be str or os.FilePath
            self.tee = open(tee, 'w+b')    
        self.tee.write(response.content)
        self.tee.seek(0)

    def readable(self):
        return True

    def read(self, size=-1):
        """Read and return up to *size* bytes by calling *self.tee.read()*.

        *size* Defaults to -1. In this case,   reads and
        returns all data until EOF. 

        Returns an empty bytes object on EOF.
        """
        return self.tee.read(size)



# TODO test str() and repr() implementations

from pandasdmx.model import (
    DEFAULT_LOCALE,
    AttributeValue,
    ContentConstraint,
    ConstraintRole,
    ConstraintRoleType,
    CubeRegion,
    DataAttribute,
    DataSet,
    DataStructureDefinition,
    DataflowDefinition,
    Dimension,
    DimensionDescriptor,
    Item,
    ItemScheme,
    Key,
    Observation,
    )
import pydantic
from pytest import raises


def test_contentconstraint():
    crole = ConstraintRole(role=ConstraintRoleType['allowable'])
    cr = ContentConstraint(role=crole)
    cr.content = {DataflowDefinition()}
    cr.data_content_region = CubeRegion(included=True, member={})


def test_dataset():
    # Enumeration values can be used to initialize
    from pandasdmx.model import ActionType
    print(ActionType)
    DataSet(action=ActionType['information'])


def test_datastructuredefinition():
    dsd = DataStructureDefinition()

    # Convenience methods
    da = dsd.attribute(id='foo')
    assert isinstance(da, DataAttribute)

    d = dsd.dimension(id='baz', order=-1)
    assert isinstance(d, Dimension)

    # from_keys()
    key1 = Key(foo=1, bar=2, baz=3)
    key2 = Key(foo=4, bar=5, baz=6)
    dsd.from_keys([key1, key2])


def test_dimension():
    # Constructor
    Dimension(id='CURRENCY', order=0)


def test_dimensiondescriptor():
    # from_key()
    key1 = Key(foo=1, bar=2, baz=3)
    dd = DimensionDescriptor.from_key(key1)

    # Key in reverse order
    key2 = Key(baz=3, bar=2, foo=1)
    assert list(key1.values.keys()) == list(reversed(key2.values.keys()))
    key3 = dd.order_key(key2)
    assert list(key1.values.keys()) == list(key3.values.keys())


def test_internationalstring():
    # Constructor; the .name attribute is an InternationalString
    i = Item(id='ECB')

    # Set and get using the attribute directly
    i.name.localizations['DE'] = 'Europäische Zentralbank'
    assert i.name.localizations['DE'] == 'Europäische Zentralbank'

    # Set and get using item convenience
    i.name['FR'] = 'Banque centrale européenne'
    assert len(i.name.localizations) == 2
    assert i.name['FR'] == 'Banque centrale européenne'

    # repr() gives all localizations
    assert repr(i.name) == '\n'.join(sorted([
        'DE: Europäische Zentralbank',
        'FR: Banque centrale européenne',
        ]))

    # Setting with a string directly sets the value in the default locale
    i.name = 'European Central Bank'
    assert len(i.name.localizations) == 1
    assert i.name.localizations[DEFAULT_LOCALE] == 'European Central Bank'

    # Setting with a (locale, text) tuple
    i.name = ('FI', 'Euroopan keskuspankki')
    assert len(i.name.localizations) == 1

    # Setting with a dict()
    i.name = {'IT': 'Banca centrale europea'}
    assert len(i.name.localizations) == 1

    # Using some other type is an error
    with raises(pydantic.ValidationError):
        i.name = 123

    # Same, but in the constructor
    i2 = Item(id='ECB', name='European Central Bank')

    # str() uses the default locale
    assert str(i2.name) == 'European Central Bank'

    # Creating with name=None raises an exception…
    with raises(pydantic.ValidationError,
                match='none is not an allowed value'):
        Item(id='ECB', name=None)

    # …giving empty dict is equivalent to giving nothing
    i3 = Item(id='ECB', name={})
    assert i3.name.localizations == Item(id='ECB').name.localizations

    # Create with iterable of 2-tuples
    i4 = Item(id='ECB', name=[
        ('DE', 'Europäische Zentralbank'),
        ('FR', 'Banque centrale européenne'),
    ])
    assert i4.name['FR'] == 'Banque centrale européenne'


def test_item():
    # Add a tree of 10 items
    items = []
    for i in range(10):
        items.append(Item(id='Foo {}'.format(i)))

        if i > 0:
            items[-1].parent = items[-2]
            items[-2].child.append(items[-1])

    # __init__(parent=...)
    Item(id='Bar 1', parent=items[0])
    assert len(items[0].child) == 2

    # __init__(child=)
    bar2 = Item(id='Bar 2', child=[items[0]])

    # __contains__()
    assert items[0] in bar2
    assert items[-1] in items[0]

    # get_child()
    assert items[0].get_child('Foo 1') == items[1]

    with raises(ValueError):
        items[0].get_child('Foo 2')


def test_itemscheme():
    is0 = ItemScheme(id='is0')
    foo0 = Item(id='foo0')

    # With a single Item

    # append()
    is0.append(foo0)

    # __getattr__
    assert is0.foo0 is foo0

    # __getitem__
    assert is0['foo0'] is foo0

    # __contains__
    assert 'foo0' in is0
    assert foo0 in is0

    # __len__
    assert len(is0) == 1

    # __repr__
    assert repr(is0) == "<ItemScheme: 'is0', 1 items>"

    # __iter__
    assert all(i is foo0 for i in is0)

    # With multiple Items

    foo1 = Item(id='foo1')
    foo2 = Item(id='foo2')
    items_list = [foo0, foo1, foo2]
    items_dict = {'foo0': foo0, 'foo1': foo1, 'foo2': foo2}

    # set with a non-dict
    is0.items = items_list
    assert is0.items == items_dict

    # set with a dict
    is0.items = items_dict
    assert is0.items == items_dict

    # extend()
    is0.items = [foo0]
    is0.extend(items_list)
    assert is0.items == items_dict

    # setdefault()
    bar0 = is0.setdefault(id='bar')
    assert bar0.id == 'bar'

    with raises(ValueError):
        is0.setdefault(foo0, id='bar')

    is0.setdefault(id='bar1', parent='foo0')
    is0.setdefault(id='bar1', parent=foo0)


def test_key():
    # Construct with a dict
    k1 = Key({'foo': 1, 'bar': 2})

    # Construct with kwargs
    k2 = Key(foo=1, bar=2)

    # Results are __eq__ each other
    assert k1 == k2

    # Doing both is an error
    with raises(ValueError):
        Key({'foo': 1}, bar=2)

    # __len__
    assert len(k1) == 2

    # __contains__: symmetrical if keys are identical
    assert k1 in k2
    assert k2 in k1
    assert Key(foo=1) in k1
    assert k1 not in Key(foo=1)

    # Set and get using item convenience
    k1['baz'] = 3  # bare value is converted to a KeyValue
    assert k1['foo'] == 1

    # __str__
    assert str(k1) == '(foo=1, bar=2, baz=3)'

    # copying: returns a new object equal to the old one
    k2 = k1.copy()
    assert id(k1) != id(k2) and k1 == k2
    # copy with changes
    k2 = Key(foo=1, bar=2).copy(baz=3)
    assert id(k1) != id(k2) and k1 == k2

    # __add__: Key with something else
    with raises(NotImplementedError):
        k1 + 4
    # Two Keys
    k2 = Key(foo=1) + Key(bar=2)
    assert k2 == k1

    # __radd__: adding a Key to None produces a Key
    assert None + k1 == k1
    # anything else is an error
    with raises(NotImplementedError):
        4 + k1

    # get_values(): preserve ordering
    assert k1.get_values() == (1, 2, 3)


def test_observation():
    obs = Observation()

    # Set by item name
    obs.attached_attribute['TIME_PERIOD'] = 3
    # NB the following does not work; see Observation.attrib()
    # obs.attrib['TIME_PERIOD'] = 3

    obs.attached_attribute['CURRENCY'] = 'USD'

    # Access by attribute name
    assert obs.attrib.TIME_PERIOD == 3
    assert obs.attrib.CURRENCY == 'USD'

    # Access by item index
    assert obs.attrib[1] == 'USD'

    # Add attributes
    obs.attached_attribute['FOO'] = '1'
    obs.attached_attribute['BAR'] = '2'
    assert obs.attrib.FOO == '1' and obs.attrib['BAR'] == '2'

    # Using classes
    da = DataAttribute(id='FOO')
    av = AttributeValue(value_for=da, value='baz')
    obs.attached_attribute[da.id] = av
    assert obs.attrib[da.id] == 'baz'



from . import Source as BaseSource


class Source(BaseSource):
    _id = 'SGR'

    def handle_response(self, response, content):
        """SGR responses do not specify content-type; set it directly."""
        if response.headers.get('content-type', None) is None:
            response.headers['content-type'] = 'application/xml'
        return response, content

    def modify_request_args(self, kwargs):
        """SGR is a data source but not a data provider.

        Override the ``agency`` argument by setting ``agency='all'`` to
        retrieve all data republished by SGR from different providers.
        """
        super().modify_request_args(kwargs)

        kwargs.setdefault('provider', 'all')
